# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""This module defines API exceptions."""


class ApiError(Exception):
    """Exception raised for API HTTP response status codes not in [200...300).

    The exception value will be the response object returned by :mod:`requests`
    which provides access to all its attributes, eg. :attr:`status_code`,
    :attr:`reason` and :attr:`text`, etc.
    """

    def __str__(self):
        """Return a string from the HTTP response causing the exception.

        The string simply lists the repsonse's status code, reason and text
        content, separated with commas.
        """
        resp = self.args[0]
        return f"{resp.status_code}, {resp.reason}, {resp.text}"


class ConfigException(Exception):
    """
    This ``ConfigException`` is raised whenever there is any error related to
    platform configuration.
    """


class AuthenticationException(Exception):
    """
    This ``AuthenticationException`` is raised either authentication
    or authorization on the platform fails.
    """

    def __init__(self, resp):
        """
        Instantiate AuthenticationException .
        :param resp: response detail will be stored in this param
        """

        self.resp = resp

    def __str__(self) -> str:
        """
        Return the message to be raised for this exception.

        :return: error message
        """
        return """An error occurred during authentication or authorization with HERE
                    platform: Status {status} -
                    Reason {reason}\n Response: {body}""".format(
            status=self.resp.status_code,
            reason=self.resp.reason,
            body=self.resp.text,
        )


class TooManyRequestsException(Exception):
    """Exception raised for API HTTP response status code 429.

    The exception value will be the response object returned by :mod:`requests`
    which provides access to all its attributes, eg. :attr:`status_code`,
    :attr:`reason` and :attr:`text`, etc.
    """

    def __init__(self, resp):
        """
        Instantiate AuthenticationException .
        :param resp: response detail will be stored in this param
        """

        self.resp = resp

    def __str__(self):
        """Return a string from the HTTP response causing the exception.

        The string simply lists the response status code, reason and text
        content, separated with commas.
        """

        return "TooManyRequestsException: Status \
                {status} - Reason {reason}\n\n" "Response: {body}".format(
            status=self.resp.status_code,
            reason=self.resp.reason,
            body=self.resp.text,
        )



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""This module defines all the configs which will be required as inputs to
    Destination Weather API."""

from .base_config import Bunch


class DestWeatherProduct(Bunch):
    """A class to define constant attributes for ``product`` parameter identifying the
        type of report to obtain.

    ``observation``:
    current weather conditions from the eight closest locations to the specified location

    ``forecast_7days``:
    morning, afternoon, evening and night weather forecasts for the next seven days.

    ``forecast_7days_simple``:
    daily weather forecasts for the next seven days

    ``forecast_hourly``:
    hourly weather forecasts for the next seven days

    ``alerts``:
    forecasted weather alerts for the next 24 hours

    ``nws_alerts``:
    all active watches and warnings for the US and Canada
    """


product = {
    "observation": "observation",
    "forecast7days": "forecast7days",
    "forecast7daysSimple": "forecast7daysSimple",
    "forecastHourly": "forecastHourly",
    "alerts": "alerts",
    "nwsAlerts": "nwsAlerts",
}

#: Use this config for ``products``` of Destination Weather API.
#: Example: for ``forecastHourly`` product use ``DEST_WEATHER_PRODUCT.forecastHourly``.
DEST_WEATHER_PRODUCT = DestWeatherProduct(**product)


class DestWeatherUnits(Bunch):
    """A class to define constant attributes for ``units`` parameter identifying
        units of measurement used.

    ``metric``:
    Follow metric system of measurements. Default.

    ``imperial``:
    Follow imperial system of measurements

    """


units = {
    "metric": "metric",
    "imperial": "imperial",
}

#: Use this config for ``units``` of Destination Weather API.
#: Example: for ``metric`` units use ``DEST_WEATHER_UNITS.metric``.
DEST_WEATHER_UNITS = DestWeatherUnits(**units)


class WeatherSeverity(Bunch):
    """A class to define the severity of the weather event

    ``insignificant``:
    Event doesn't have significance by nature

    ``no_alerts``:
    There are no alerts for the location

    ``minor``:
    Minor Severity, the event is potentially dangerous but not usual

    ``medium``:
    Medium Severity, the event is dangerous

    ``high``:
    High Severity, The event is very dangerous

    ``emergency``:
    Emergency. Take immediate action to protect life.
    """


weather_severity = {
    "insignificant": 0,
    "no_alerts": 1,
    "minor": 2,
    "medium": 3,
    "high": 4,
    "emergency": 5,
}

#: Use this config for ``weather_severity``` of get weather alerts endpoint.
#: Example: for ``high severity`` events use ``WEATHER_SEVERITY.high``.
WEATHER_SEVERITY = WeatherSeverity(**weather_severity)


class WeatherType(Bunch):
    """A class to define the type of the weather event"""


weather_type = {
    "extremely_high_temperature": 1,
    "extremely_low_temperature": 2,
    "fog": 3,
    "ice": 4,
    "rain": 5,
    "snow": 6,
    "thunderstorm": 7,
    "wind": 8,
    "air_quality": 9,
    "volcanic_ashfall": 10,
    "avalanche": 11,
    "tsunami": 12,
    "dust_storm": 13,
    "earthquake": 14,
    "fire_danger": 15,
    "flood": 16,
    "high_waves": 17,
    "gigh_uv_index": 18,
    "low_water": 19,
    "smoke": 20,
    "volcano": 21,
    "ice_in_waterway": 22,
    "coastal_event": 23,
    "civil_danger": 24,
    "evacuation": 25,
    "hazardous_material": 26,
    "radiological_hazard": 27,
    "shelter_in_place": 28,
    "warning": 29,
}

#: Use this config for ``weather_type``` of get weather alerts endpoint of Destination Weather API.
#: Example: for ``fog`` weather type use ``WEATHER_TYPE.fog``.
WEATHER_TYPE = WeatherType(**weather_type)



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""Module for testing API exceptions."""


import pytest

from here_location_services import LS
from here_location_services.apis import Api as BaseApi
from here_location_services.exceptions import (
    ApiError,
    AuthenticationException,
    TooManyRequestsException,
)
from here_location_services.platform.apis.api import Api
from tests.conftest import get_mock_response


def test_exception_requests_inalid():
    """Raise exception via requests as response for invalid api_key."""
    ls = LS(api_key="dummy")
    with pytest.raises(ApiError) as execinfo:
        resp = ls.geocode(query="abc")
        raise ApiError(resp)
    resp = execinfo.value.args[0]
    assert resp.status_code == 401
    assert resp.reason == "Unauthorized"
    assert str(execinfo.value).startswith('401, Unauthorized, {"error":"Unauthorized"')


def test_raise_response_exception():
    reason = "This is mock reason"
    text = "This is mock text"
    mock_response = get_mock_response(401, reason, text)
    with pytest.raises(AuthenticationException) as execinfo:
        Api.raise_response_exception(mock_response)
    resp = execinfo.value.args[0]
    assert resp.status_code == 401
    assert resp.reason == "This is mock reason"
    assert "This is mock text" in (str(execinfo.value))
    mock_response = get_mock_response(429, reason, text)
    with pytest.raises(TooManyRequestsException) as execinfo:
        Api.raise_response_exception(mock_response)
    resp = execinfo.value.args[0]
    assert resp.status_code == 429
    assert resp.reason == "This is mock reason"
    assert "This is mock text" in (str(execinfo.value))
    mock_response = get_mock_response(500, reason, text)
    with pytest.raises(Exception):
        Api.raise_response_exception(mock_response)


def test_auth_exception():
    """Test if both api key and auth token in not provided."""
    api = BaseApi(api_key=None, auth=None)
    with pytest.raises(Exception):
        api._get_url_string()



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

import os
from collections import namedtuple

import pytest

from here_location_services.autosuggest_api import AutosuggestApi
from here_location_services.destination_weather_api import DestinationWeatherApi
from here_location_services.exceptions import ConfigException
from here_location_services.geocoding_search_api import GeocodingSearchApi
from here_location_services.isoline_routing_api import IsolineRoutingApi
from here_location_services.platform.credentials import PlatformCredentials
from here_location_services.routing_api import RoutingApi
from here_location_services.utils import get_apikey

LS_API_KEY = get_apikey()
HERE_USER_ID = os.environ.get("HERE_USER_ID")
HERE_CLIENT_ID = os.environ.get("HERE_CLIENT_ID")
HERE_ACCESS_KEY_ID = os.environ.get("HERE_ACCESS_KEY_ID")
HERE_ACCESS_KEY_SECRET = os.environ.get("HERE_ACCESS_KEY_SECRET")


@pytest.fixture()
def geo_search_api():
    """Create shared low level Location services GeocodingSearchApi instance as a pytest fixture.
    :return: :class:`GeocodingSearchApi` object.
    """
    api = GeocodingSearchApi(api_key=LS_API_KEY)
    return api


@pytest.fixture()
def isoline_routing_api():
    """Create shared low level Location services IsolineRoutingApi instance as a pytest fixture."""
    api = IsolineRoutingApi(api_key=LS_API_KEY)
    return api


@pytest.fixture()
def routing_api():
    """Create shared low level Location services IsolineRoutingApi instance as a pytest fixture."""
    api = RoutingApi(api_key=LS_API_KEY)
    return api


@pytest.fixture()
def autosuggest_api():
    """Create shared low level Location services AutosuggestApi instance as a pytest fixture."""
    api = AutosuggestApi(api_key=LS_API_KEY)
    return api


@pytest.fixture()
def destination_weather_api():
    """Create shared low level Location services Dest Weather instance as a pytest fixture."""
    api = DestinationWeatherApi(api_key=LS_API_KEY)
    return api


def env_setup_done():
    """This function will return env variables required to authenticate APIs using auth token."""
    try:
        _ = PlatformCredentials.from_default()
    except (ConfigException, AssertionError):
        return False
    else:
        return True


def get_mock_response(status_code: int, reason: str, text: str):
    """
    Return mock response.

    :param status_code: An int representing status_code.
    :param reason: A string to represent reason.
    :param text: A string to represent text.
    :return: MockResponse object.
    """
    MockResponse = namedtuple("MockResponse", ["status_code", "reason", "text"])
    mock_response = MockResponse(status_code, reason, text)
    return mock_response



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains base classes for accessing the Location Services RESTful APIs.
"""

import urllib
import urllib.request
from typing import Dict, Optional

import requests

from here_location_services.config.url_config import conf
from here_location_services.platform.auth import Auth


class Api:
    """A base class for low-level HTTP RESTful API client for location services."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        auth: Optional[Auth] = None,
        proxies: Optional[dict] = None,
        country: str = "row",
    ):
        self.auth = auth
        self.credentials = dict(
            api_key=api_key, access_token=self.auth.token if self.auth else None
        )
        self.proxies = proxies or urllib.request.getproxies()
        self.headers: Dict[str, str] = {}
        self.country = country

    def _get_url_string(self) -> str:
        """
        Get url string from config based on type of country.

        For china url string ends with ``hereapi.cn`` and for rest of the countries
        deonoted by ``row`` it is ``hereapi.com``.

        :raises Exception: If ``api_key`` not found in credentials.
        """
        if self.credentials["api_key"] or self.credentials["access_token"]:
            url = conf[self.country]["here_api"]
            return url
        else:
            raise Exception(
                f"api_key: {self.credentials['api_key']} is not present in credentials."
            )

    def __add_api_key_in_params(self, params: Dict) -> Dict:
        """
        Add api_key in query params dictionary.

        :return: Dict.
        """
        params.update({"apiKey": self.credentials["api_key"]})
        return params

    def get(self, url: str, params: Optional[Dict] = None, **kwargs):
        """Send HTTP GET request.

        :param url: A string to represent URL.
        :param params: An optional dict for query params.
        :param kwargs: An optional extra arguments.
        :return: :class:`requests.Response` object.
        """
        q_params = params if params is not None else {}
        if self.credentials["api_key"]:
            q_params = self.__add_api_key_in_params(q_params)
        elif self.credentials["access_token"]:
            auth_token = {"Authorization": f"Bearer {self.credentials['access_token']}"}
            self.headers.update(auth_token)
        resp = requests.get(url, params=q_params, headers=self.headers, **kwargs)
        return resp

    def post(self, url: str, data: Dict, params: Optional[Dict] = None):
        """
        Send HTTP POST request.

        :param url: A string to represent URL.
        :param data: A dictionary to represent the post data
        :param params: An optional dict for query params.
        :return: :class:`requests.Response` object.
        """
        self.headers.update({"Content-Type": "application/json"})
        q_params = params if params is not None else {}
        if self.credentials["api_key"]:
            q_params = self.__add_api_key_in_params(q_params)
        elif self.credentials["access_token"]:
            auth_token = {"Authorization": f"Bearer {self.credentials['access_token']}"}
            self.headers.update(auth_token)
        resp = requests.post(
            url, params=q_params, json=data, proxies=self.proxies, headers=self.headers
        )
        return resp



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0
"""
This module contains an :class:`AAAOauth2ApiClient` class to perform oauth API operations.

The HERE API reference documentation used in this module can be found here:
|iam_api_reference|

.. |iam_api_reference| raw:: html

   <a href="https://developer.here.com/documentation/identity-access-management/api-reference-swagger.html">IAM API Reference</a>
"""  # noqa

from typing import Dict, Optional

from requests_oauthlib import OAuth1

from here_location_services.platform.apis.api import Api


class AAAOauth2Api(Api):
    """
    This class provides access to HERE platform AAA Oauth2 APIs.
    """

    def __init__(
        self,
        base_url: str,
        proxies: Optional[dict] = None,
    ):
        self.base_url = base_url
        self.proxies: Optional[Dict] = proxies
        super().__init__(
            access_token=None,
            proxies=self.proxies,
        )

    def request_scoped_access_token(self, oauth: OAuth1, data: str) -> Dict:  # type: ignore[return]  # noqa E501
        """
        Request scoped access oauth2 token from platform.

        :param oauth: oauth1 configuration.
        :param data: a string which represents request body.
        :return: a json with scoped access token.
        """
        resp = self.post(
            url=self.base_url,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data=data,
            auth=oauth,
        )
        if resp.status_code == 200:
            resp_dict: dict = resp.json()
            return resp_dict
        else:
            self.raise_response_exception(resp)



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

from argparse import Namespace
from datetime import datetime

import pytest
import requests
from geojson import Point

from here_location_services.config.dest_weather_config import DEST_WEATHER_PRODUCT
from here_location_services.config.isoline_routing_config import (
    ISOLINE_ROUTING_TRANSPORT_MODE,
    RANGE_TYPE,
)
from here_location_services.config.matrix_routing_config import WorldRegion
from here_location_services.exceptions import ApiError
from here_location_services.matrix_routing_api import MatrixRoutingApi
from here_location_services.utils import get_apikey

LS_API_KEY = get_apikey()


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_destination_weather(destination_weather_api):
    """Test Destination Weather api."""
    resp = destination_weather_api.get_dest_weather(
        products=[DEST_WEATHER_PRODUCT.observation], query="Chicago"
    )
    assert type(resp) == requests.Response
    assert resp.status_code == 200


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_weather_alerts(destination_weather_api):
    """Test Destination Weather api."""
    resp = destination_weather_api.get_weather_alerts(
        geometry=Point(coordinates=[15.256, 23.456]),
        start_time=datetime.now(),
    )
    assert type(resp) == requests.Response
    assert resp.status_code == 200


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_autosuggest(autosuggest_api):
    """Test autosuggest api."""
    resp = autosuggest_api.get_autosuggest(query="res", limit=5, at=["52.93175,12.77165"])
    assert type(resp) == requests.Response
    assert resp.status_code == 200


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_geocoding(geo_search_api):
    """Test geocoding api."""
    address = "Goregaon West, Mumbai 400062, India"
    resp = geo_search_api.get_geocoding(query=address)
    assert type(resp) == requests.Response
    assert resp.status_code == 200


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_reverse_geocoding(geo_search_api):
    """Test reverse geocoding."""
    resp = geo_search_api.get_reverse_geocoding(lat=19.1646, lng=72.8493)
    assert type(resp) == requests.Response
    assert resp.status_code == 200


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_isonline_routing(isoline_routing_api):
    """Test isonline routing api."""
    result = isoline_routing_api.get_isoline_routing(
        origin=[52.5, 13.4],
        range="3000",
        range_type=RANGE_TYPE.distance,
        transport_mode=ISOLINE_ROUTING_TRANSPORT_MODE.car,
    )

    coordinates = result.json()["isolines"][0]["polygons"][0]["outer"]
    assert coordinates[0]


def test_mock_api_error(mocker):
    """Mock Test for geocoding api."""
    mock_response = Namespace(status_code=300)
    mocker.patch(
        "here_location_services.matrix_routing_api.requests.post",
        return_value=mock_response,
    )
    origins = [
        {"lat": 37.76, "lng": -122.42},
        {"lat": 40.63, "lng": -74.09},
        {"lat": 30.26, "lng": -97.74},
    ]
    region_definition = WorldRegion()
    mat = MatrixRoutingApi(api_key="dummy")
    with pytest.raises(ApiError):
        _ = mat.matrix_route(
            origins=origins,
            region_definition=region_definition,
        )



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""This module contains classes for accessing `HERE Routing API <https://developer.here.com/documentation/routing-api/8.17.0/dev_guide/index.html>`_.
"""  # noqa E501

from datetime import datetime
from typing import Dict, List, Optional

from here_location_services.config.base_config import PlaceOptions, Truck, WayPointOptions
from here_location_services.config.matrix_routing_config import AvoidBoundingBox
from here_location_services.config.routing_config import Scooter, Via
from here_location_services.platform.auth import Auth

from .apis import Api
from .exceptions import ApiError


class RoutingApi(Api):
    """A class for accessing HERE routing APIs."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        auth: Optional[Auth] = None,
        proxies: Optional[dict] = None,
        country: str = "row",
    ):
        super().__init__(api_key, auth=auth, proxies=proxies, country=country)
        self._base_url = f"https://router.{self._get_url_string()}"

    def route(
        self,
        transport_mode: str,
        origin: List,
        destination: List,
        via: Optional[List[Via]] = None,
        origin_place_options: Optional[PlaceOptions] = None,
        origin_waypoint_options: Optional[WayPointOptions] = None,
        destination_place_options: Optional[PlaceOptions] = None,
        destination_waypoint_options: Optional[WayPointOptions] = None,
        scooter: Optional[Scooter] = None,
        departure_time: Optional[datetime] = None,
        routing_mode: str = "fast",
        alternatives: int = 0,
        units: str = "metric",
        lang: str = "en-US",
        return_results: Optional[List] = None,
        spans: Optional[List] = None,
        truck: Optional[Truck] = None,
        avoid_features: Optional[List[str]] = None,
        avoid_areas: Optional[List[AvoidBoundingBox]] = None,
        exclude: Optional[List[str]] = None,
    ):
        """Calculate route between two endpoints.

        See further information `here <https://developer.here.com/documentation/routing-api/8.16.0/api-reference-swagger.html>_`.

        :param transport_mode: A string to represent mode of transport.
        :param origin: A list of ``latitude`` and ``longitude`` of origin point of route.
        :param destination: A list of ``latitude`` and ``longitude`` of destination point of route.
        :param via: A list of tuples of ``latitude`` and ``longitude`` of via points.
        :param origin_place_options: :class:`PlaceOptions` optinal place options for ``origin``.
        :param origin_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``origin``.
        :param destination_place_options: :class:`PlaceOptions` optinal place options
            for ``destination``.
        :param destination_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``destination``.
        :param scooter: Additional attributes for scooter route.
        :param departure_time: :class:`datetime.datetime` object.
        :param routing_mode: A string to represent routing mode.
        :param alternatives: Number of alternative routes to return aside from the optimal route.
            default value is ``0`` and maximum is ``6``.
        :param units: A string representing units of measurement used in guidance instructions.
            The default is metric.
        :param lang: A string representing preferred language of the response.
            The value should comply with the IETF BCP 47.
        :param return_results: A list of strings.
        :param spans: A list of strings to define which attributes are included in the response
            spans.
        :param truck: Different truck options to use during route calculation.
            use object of :class:`Truck here_location_services.config.matrix_routing_config.Truck>`
        :param avoid_features: Avoid routes that violate these properties. Avoid features are
            defined in :attr:`AVOID_FEATURES <here_location_services.config.routing_config.AVOID_FEATURES>`
        :param avoid_areas: A list of areas to avoid during route calculation. To define avoid area
            use object of :class:`AvoidBoundingBox here_location_services.config.matrix_routing_config.AvoidBoundingBox>`.
        :param exclude: A comma separated list of three-letter country codes
            (ISO-3166-1 alpha-3 code) that routes will exclude.
        :return: :class:`requests.Response` object.
        :raises ApiError: If ``status_code`` of API response is not 200.
        """  # noqa E501
        path = "v8/routes"
        url = f"{self._base_url}/{path}"
        params: Dict[str, str] = {
            "origin": ",".join([str(i) for i in origin]),
            "destination": ",".join([str(i) for i in destination]),
            "transportMode": transport_mode,
        }
        if via:
            vias = []
            for v in via:
                via_str = f"via={v.lat},{v.lng}"
                if v.place_options:
                    place_optns_str = ";".join(
                        key + "=" + str(val)
                        for key, val in vars(v.place_options).items()
                        if val is not None
                    )
                    via_str = via_str + ";" + place_optns_str
                if v.waypoint_options:
                    waypoint_optns_str = "!".join(
                        key + "=" + str(val)
                        for key, val in vars(v.waypoint_options).items()
                        if val is not None
                    )
                    via_str = via_str + "!" + waypoint_optns_str
                vias.append(via_str)
            url = url + "?" + "&".join(vias)
        if departure_time:
            params["departureTime"] = departure_time.isoformat(timespec="seconds")
        params["routingMode"] = routing_mode
        params["alternatives"] = str(alternatives)
        params["units"] = units
        params["lang"] = lang
        if return_results:
            params["return"] = ",".join(return_results)
        if spans:
            params["spans"] = ",".join(spans)

        if origin_place_options:
            origin_place_opt = ";".join(
                key + "=" + str(val)
                for key, val in vars(origin_place_options).items()
                if val is not None
            )
            params["origin"] = ";".join([params["origin"], origin_place_opt])

        if origin_waypoint_options:
            origin_way_opt = "!".join(
                key + "=" + str(val)
                for key, val in vars(origin_waypoint_options).items()
                if val is not None
            )
            params["origin"] = "!".join([params["origin"], origin_way_opt])

        if destination_place_options:
            dest_place_opt = ";".join(
                key + "=" + str(val)
                for key, val in vars(destination_place_options).items()
                if val is not None
            )
            params["destination"] = ";".join([params["destination"], dest_place_opt])

        if destination_waypoint_options:
            dest_way_opt = "!".join(
                key + "=" + str(val)
                for key, val in vars(destination_waypoint_options).items()
                if val is not None
            )
            params["destination"] = "!".join([params["destination"], dest_way_opt])

        if scooter:
            if scooter.allowHighway is True:
                params["scooter[allowHighway]"] = "true"
            else:
                params["scooter[allowHighway]"] = "false"

        if truck:
            for key, val in vars(truck).items():
                if key == "shippedHazardousGoods" and val:
                    params[f"truck[{key}]"] = ",".join(val)
                elif val is not None:
                    params[f"truck[{key}]"] = val

        if avoid_features:
            params["avoid[features]"] = ",".join(avoid_features)

        if avoid_areas:
            bbox_areas = []
            for area in avoid_areas:
                area_attrs = vars(area)
                bbox_areas.append(
                    "bbox:{west},{south},{east},{north}".format(
                        west=area_attrs["west"],
                        south=area_attrs["south"],
                        east=area_attrs["east"],
                        north=area_attrs["north"],
                    )
                )
            params["avoid[areas]"] = "|".join(bbox_areas)

        if exclude:
            params["exclude"] = ",".join(exclude)
        resp = self.get(url, params=params, proxies=self.proxies)
        if resp.status_code == 200:
            return resp
        else:
            raise ApiError(resp)



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

import os

from here_location_services.utils import get_apikey


def test_get_apikey():
    """Test ``get_apikey`` returns empty string when
    environment variable ``LS_API_KEY`` is not set.
    """
    api_key = os.environ.get("LS_API_KEY")
    del os.environ["LS_API_KEY"]
    key = get_apikey()
    assert key == ""
    os.environ["LS_API_KEY"] = api_key



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""This module defines classes to manage Platform credentials."""
from os import getenv
from os.path import expanduser, expandvars
from pathlib import Path
from typing import List, Union

from pyhocon import ConfigFactory, ConfigMissingException, ConfigTree

from here_location_services.exceptions import ConfigException

DEFAULT_CREDENTIALS_PATH = "~/.here/credentials.properties"


class PlatformCredentials:
    """
    Credentials provides functions for dealing with the HERE platform
    Credentials.

    Credentials can be read from the following locations:

    - The default location: "~/.here/credentials.properties"
    - A custom path to a credentials properties file
    - Environment variables
    """

    def __init__(
        self,
        cred_properties: ConfigTree,
    ):
        """
        Instantiate the credentials object.

        :param cred_properties: the properties of Credentials.
        """
        self.cred_properties = cred_properties

    @classmethod
    def from_default(cls) -> "PlatformCredentials":
        """Return the credentials object from the default default credential
        path at '~/.here/credentials.properties'.

        If environmental variables are set, these values will override the ones
        found in the default file.

        If no default file is found, this method will try to read the
        credentials from the environmental variables.

        :return: credentials
        """

        try:
            credentials = cls.from_credentials_file(DEFAULT_CREDENTIALS_PATH)
        except (ConfigException, FileNotFoundError):
            credentials = cls.from_env()

        credentials.patch_using_env()
        return credentials

    @classmethod
    def from_credentials_file(cls, path: Union[str, Path]) -> "PlatformCredentials":
        """
        Return the credentials object from a specified credentials path.

        :param path: path to a HERE platform credentials.properties file.
        :return: credentials
        :raises ConfigException: Erroneous credentials.properties file in path

        """
        credentials_path = expanduser(expandvars(path))
        try:
            credentials_properties = ConfigFactory.parse_file(credentials_path)
            user = credentials_properties.get("here.user.id")
            client = credentials_properties["here.client.id"]
            key = credentials_properties["here.access.key.id"]
            secret = credentials_properties["here.access.key.secret"]
            endpoint = credentials_properties["here.token.endpoint.url"]
            if user and client and key and secret and endpoint:
                credentials_config = ConfigTree()
                credentials_config.put("user", user)
                credentials_config.put("client", client)
                credentials_config.put("key", key)
                credentials_config.put("secret", secret)
                credentials_config.put("endpoint", endpoint)
                return PlatformCredentials(credentials_config)
            else:
                raise ConfigException("Erroneous ", credentials_path, " file")
        except ConfigMissingException:
            raise ConfigException("Erroneous ", credentials_path, " file")

    @classmethod
    def from_env(cls) -> "PlatformCredentials":
        """
        Return the credentials object from the following environment variables:

          - ``HERE_USER_ID``
          - ``HERE_CLIENT_ID``
          - ``HERE_ACCESS_KEY_ID``
          - ``HERE_ACCESS_KEY_SECRET``
          - ``HERE_TOKEN_ENDPOINT_URL`` (optional)

        :return: credentials parsed from the environment variables
        :raises ConfigException: missing environmental variables that are mandatory
        """

        user = getenv("HERE_USER_ID")
        client = getenv("HERE_CLIENT_ID")
        access_key_id = getenv("HERE_ACCESS_KEY_ID") or getenv("HERE_ACCESS_KEY")
        access_key_secret = getenv("HERE_ACCESS_KEY_SECRET") or getenv("HERE_ACCESS_SECRET")
        endpoint = (
            getenv("HERE_TOKEN_ENDPOINT_URL")
            or getenv("HERE_TOKEN_ENDPOINT")
            or "https://account.api.here.com/oauth2/token"
        )

        missing_env_vars: List[str] = []

        if not user:
            missing_env_vars.append("HERE_USER_ID")
        if not client:
            missing_env_vars.append("HERE_CLIENT_ID")
        if not access_key_id:
            missing_env_vars.append("HERE_ACCESS_KEY_ID")
        if not access_key_secret:
            missing_env_vars.append("HERE_ACCESS_KEY_SECRET")

        if missing_env_vars:
            raise ConfigException(
                "Missing environmental variables: {}".format(", ".join(missing_env_vars))
            )

        # at this points, we should have all the variables with a non-empty value
        assert user and client and access_key_id and access_key_secret and endpoint

        credentials_config = ConfigTree()
        credentials_config.put("user", user)
        credentials_config.put("client", client)
        credentials_config.put("key", access_key_id)
        credentials_config.put("secret", access_key_secret)
        credentials_config.put("endpoint", endpoint)
        return cls(credentials_config)

    def patch_using_env(self):
        """
        Patch the credentials by reading the following environment variables and
        applying them accordingly.

          - ``HERE_USER_ID``
          - ``HERE_CLIENT_ID``
          - ``HERE_ACCESS_KEY_ID``
          - ``HERE_ACCESS_KEY_SECRET``
          - ``HERE_TOKEN_ENDPOINT_URL``

        Whenever such an environment variable is set,
        it overrides the one loaded from file.
        """
        if self.cred_properties:
            credentials_config = self.cred_properties

            user = getenv("HERE_USER_ID") or credentials_config["user"]
            client = getenv("HERE_CLIENT_ID") or credentials_config["client"]
            key = (
                getenv("HERE_ACCESS_KEY_ID")
                or getenv("HERE_ACCESS_KEY")
                or credentials_config["key"]
            )
            secret = (
                getenv("HERE_ACCESS_KEY_SECRET")
                or getenv("HERE_ACCESS_SECRET")
                or credentials_config["secret"]
            )
            endpoint = (
                getenv("HERE_TOKEN_ENDPOINT_URL")
                or getenv("HERE_TOKEN_ENDPOINT")
                or credentials_config["endpoint"]
            )
            credentials_config["user"] = user
            credentials_config["client"] = client
            credentials_config["key"] = key
            credentials_config["secret"] = secret
            credentials_config["endpoint"] = endpoint



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0
"""This module will test platform credentials."""

import tempfile
from pathlib import Path

import pytest

from here_location_services.exceptions import ConfigException
from here_location_services.platform.credentials import PlatformCredentials


def test_from_credentials_file():
    """Test credentials from file."""
    file_path = Path(__file__).parent / Path("data") / Path("dummy_credentials.properties")
    cred = PlatformCredentials.from_credentials_file(file_path)
    assert cred.cred_properties["user"] == "dummy_user_id"
    assert cred.cred_properties["client"] == "dummy_client_id"
    assert cred.cred_properties["key"] == "dummy_access_key_id"
    assert cred.cred_properties["secret"] == "dummy_access_key_secret"
    assert cred.cred_properties["endpoint"] == "dummy_token_endpoint"
    with pytest.raises(ConfigException):
        with tempfile.NamedTemporaryFile() as tmp:
            _ = PlatformCredentials.from_credentials_file(tmp.name)


def test_erronous_credential_file():
    """Test credential file with error."""
    file_path = Path(__file__).parent / Path("data") / Path("erroneous_credential.properties")
    with pytest.raises(ConfigException) as execinfo:
        PlatformCredentials.from_credentials_file(file_path)
    assert execinfo.value.args[0] == "Erroneous "



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""This module contains classes for accessing `HERE Routing API <https://developer.here.com/documentation/routing/dev_guide/topics/request-isoline.html>`_.
"""  # noqa E501
from datetime import datetime
from typing import Any, Dict, List, Optional

import requests

from here_location_services.platform.auth import Auth

from .apis import Api
from .config.base_config import PlaceOptions, Truck, WayPointOptions
from .exceptions import ApiError


class IsolineRoutingApi(Api):
    """A class for accessing HERE isoline routing API."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        auth: Optional[Auth] = None,
        proxies: Optional[dict] = None,
        country: str = "row",
    ):
        super().__init__(api_key, auth=auth, proxies=proxies, country=country)
        self._base_url = f"https://isoline.router.{self._get_url_string()}"

    def get_isoline_routing(
        self,
        range: str,
        range_type: str,
        transport_mode: str,
        origin: Optional[List] = None,
        departure_time: Optional[datetime] = None,
        destination: Optional[List] = None,
        arrival_time: Optional[datetime] = None,
        routing_mode: Optional[str] = "fast",
        shape_max_points: Optional[int] = None,
        optimised_for: Optional[str] = "balanced",
        avoid_features: Optional[List[str]] = None,
        truck: Optional[Truck] = None,
        origin_place_options: Optional[PlaceOptions] = None,
        origin_waypoint_options: Optional[WayPointOptions] = None,
        destination_place_options: Optional[PlaceOptions] = None,
        destination_waypoint_options: Optional[WayPointOptions] = None,
    ) -> requests.Response:
        """Get isoline routing.

        Request a polyline that connects the endpoints of all routes
        leaving from one defined center with either a specified length
        or specified travel time.

        :param range: A string representing a range of isoline, unit is defined by
            parameter ``range_type``. Example: range='1000' or range='1000,2000,3000'
        :param range_type: A string representing a type of ``range``. Possible values are
            ``distance``, ``time`` and ``consumption``. For distance the unit meters. For a
            time the unit is seconds. For consumption, it is defined by the consumption
            model.
        :param transport_mode: A string representing Mode of transport to be used for the
            calculation of the isolines.
            Example: ``car``.
        :param origin: Center of the isoline request. The Isoline(s) will cover the region
            which can be reached from this point within given range. It cannot be used in
            combination with ``destination`` parameter.
        :param departure_time: Specifies the time of departure as defined by either date-time
            or full-date partial-time in RFC 3339, section 5.6 (for example, 2019-06-24T01:23:45).
            The requested time is converted to the local time at origin. When the optional timezone
            offset is not specified, time is assumed to be local. If neither departure_time or
            arrival_time are specified, current time at departure location will be used. All Time
            values in the response are returned in the timezone of each location.
        :param destination: Center of the isoline request. The Isoline(s) will cover the
            region within the specified range that can reach this point. It cannot be used
            in combination with ``origin`` parameter.
        :param arrival_time: Specifies the time of arrival as defined by either date-time or
            full-date T partial-time in RFC 3339, section 5.6 (for example, 2019-06-24T01:23:45).
            The requested time is converted to the local time at destination. When the optional
            timezone offset is not specified, time is assumed to be local. All Time values in
            the response are returned in the timezone of each location.
        :param routing_mode: A string to represent routing mode.
        :param shape_max_points: An integer to Limit the number of points in the resulting isoline
            geometry. If the isoline consists of multiple components, the sum of points from all
            components is considered. This parameter doesn't affect performance.
        :param optimised_for: A string to specify how isoline calculation is optimized.
        :param avoid_features: Avoid routes that violate these properties. Avoid features
            are defined in :attr:
            `AVOID_FEATURES <here_location_services.config.isoline_routing_config.AVOID_FEATURES>`
        :param truck: Different truck options to use during route calculation when transport_mode
            = truck. use object of :class:`Truck here_location_services.config.base_config.Truck>`
        :param origin_place_options: :class:`PlaceOptions` optinal place options for ``origin``.
        :param origin_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``origin``.
        :param destination_place_options: :class:`PlaceOptions` optinal place options
            for ``destination``.
        :param destination_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``destination``.
        :return: :class:`requests.Response` object.
        :raises ApiError: If ``status_code`` of API response is not 200.
        """
        path = "v8/isolines"
        url = f"{self._base_url}/{path}"
        params: Dict[str, Any] = {
            "range[type]": range_type,
            "range[values]": range,
            "transportMode": transport_mode,
        }
        if origin:
            params["origin"] = ",".join([str(i) for i in origin])
        if destination:
            params["destination"] = ",".join([str(i) for i in destination])
        if arrival_time:
            params["arrivalTime"] = arrival_time.isoformat(timespec="seconds")
        if departure_time:
            params["departureTime"] = departure_time.isoformat(timespec="seconds")
        if routing_mode:
            params["routingMode"] = routing_mode
        if optimised_for:
            params["optimizeFor"] = optimised_for
        if avoid_features:
            avoid: Dict[str, Any] = {"features": avoid_features}
            params["avoid"] = avoid
        if truck:
            params["truck"] = {k: v for k, v in vars(truck).items() if v is not None}
        if shape_max_points:
            params["shape[maxPoints]"] = shape_max_points

        if origin_place_options:
            origin_place_opt = ";".join(
                key + "=" + str(val)
                for key, val in vars(origin_place_options).items()
                if val is not None
            )
            params["origin"] = ";".join([params["origin"], origin_place_opt])

        if origin_waypoint_options:
            origin_way_opt = "!".join(
                key + "=" + str(val)
                for key, val in vars(origin_waypoint_options).items()
                if val is not None
            )
            params["origin"] = "!".join([params["origin"], origin_way_opt])

        if destination_place_options:
            dest_place_opt = ";".join(
                key + "=" + str(val)
                for key, val in vars(destination_place_options).items()
                if val is not None
            )
            params["destination"] = ";".join([params["destination"], dest_place_opt])

        if destination_waypoint_options:
            dest_way_opt = "!".join(
                key + "=" + str(val)
                for key, val in vars(destination_waypoint_options).items()
                if val is not None
            )
            params["destination"] = "!".join([params["destination"], dest_way_opt])
        resp = self.get(url, params=params, proxies=self.proxies)
        if resp.status_code == 200:
            return resp
        else:
            raise ApiError(resp)



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains top level imports for ``here_location_services`` package.
"""

from .__version__ import __version__  # noqa: F401
from .ls import LS  # noqa: F401
from .platform.credentials import PlatformCredentials  # noqa: F401



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

import json
from datetime import datetime, timedelta

import pandas as pd
import pytest
import pytz
from geojson import FeatureCollection, Point

from here_location_services import LS
from here_location_services.config.autosuggest_config import POLITICAL_VIEW, SHOW, SearchCircle
from here_location_services.config.base_config import (
    ROUTING_MODE,
    SHIPPED_HAZARDOUS_GOODS,
    PlaceOptions,
    Truck,
    WayPointOptions,
)
from here_location_services.config.dest_weather_config import (
    DEST_WEATHER_PRODUCT,
    DEST_WEATHER_UNITS,
    WEATHER_SEVERITY,
    WEATHER_TYPE,
)
from here_location_services.config.isoline_routing_config import (
    ISOLINE_ROUTING_AVOID_FEATURES,
    ISOLINE_ROUTING_TRANSPORT_MODE,
    RANGE_TYPE,
)
from here_location_services.config.matrix_routing_config import (
    AVOID_FEATURES,
    MATRIX_ATTRIBUTES,
    PROFILE,
    AutoCircleRegion,
    AvoidBoundingBox,
    BoundingBoxRegion,
    CircleRegion,
    PolygonRegion,
    WorldRegion,
)
from here_location_services.config.routing_config import AVOID_FEATURES as ROUTING_AVOID_FEATURES
from here_location_services.config.routing_config import (
    ROUTE_COURSE,
    ROUTE_MATCH_SIDEOF_STREET,
    ROUTING_RETURN,
    ROUTING_SPANS,
    ROUTING_TRANSPORT_MODE,
    Scooter,
    Via,
)
from here_location_services.config.search_config import PLACES_CATEGORIES
from here_location_services.exceptions import ApiError
from here_location_services.responses import GeocoderResponse
from here_location_services.utils import get_apikey

LS_API_KEY = get_apikey()


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_ls_weather_alerts():
    """Test weather alerts endpoint of destination weather api."""
    ls = LS(api_key=LS_API_KEY)
    resp = ls.get_weather_alerts(
        geometry=Point(coordinates=[15.256, 23.456]),
        start_time=datetime.now(),
        width=3000,
    )
    assert resp

    resp2 = ls.get_weather_alerts(
        geometry=Point(coordinates=[15.256, 23.456]),
        start_time=datetime.now(),
        weather_type=WEATHER_TYPE.ice,
        weather_severity=WEATHER_SEVERITY.high,
        country="US",
        end_time=datetime.now() + timedelta(days=7),
    )
    assert resp2

    with pytest.raises(ApiError):
        ls2 = LS(api_key="dummy")
        ls2.get_weather_alerts(
            geometry=Point(coordinates=[15.256, 23.456]),
            start_time=datetime.now(),
        )


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_ls_dest_weather():
    """Test destination weather api."""
    ls = LS(api_key=LS_API_KEY)
    resp = ls.get_dest_weather(products=[DEST_WEATHER_PRODUCT.observation], query="Chicago")
    assert resp.places
    assert resp.places[0]["observations"]

    resp2 = ls.get_dest_weather(
        products=[DEST_WEATHER_PRODUCT.forecastHourly],
        query="Chicago",
        units=DEST_WEATHER_UNITS.imperial,
    )
    assert resp2.places
    assert resp2.places[0]["hourlyforecasts"]

    resp3 = ls.get_dest_weather(
        products=[DEST_WEATHER_PRODUCT.forecast7days], at=["-13.163068,-72.545128"]
    )
    assert resp3.places
    assert resp3.places[0]["extendedDailyforecasts"]

    resp4 = ls.get_dest_weather(
        products=[DEST_WEATHER_PRODUCT.forecast7daysSimple, DEST_WEATHER_PRODUCT.observation],
        zipcode="10025",
        one_observation=True,
    )
    assert resp4.places
    assert resp4.places[0]["observations"]

    resp5 = ls.get_dest_weather(
        products=[DEST_WEATHER_PRODUCT.forecast7daysSimple, DEST_WEATHER_PRODUCT.observation],
        zipcode="10025",
        at=["-13.163068,-72.545128"],
        one_observation=True,
    )
    assert resp5.places
    assert resp5.places[0]["observations"]

    with pytest.raises(ValueError):
        ls.get_dest_weather(products=[DEST_WEATHER_PRODUCT.forecast7days])

    with pytest.raises(ValueError):
        ls.get_dest_weather(
            products=[DEST_WEATHER_PRODUCT.forecast7days], query="Chicago", one_observation=True
        )

    with pytest.raises(ValueError):
        ls.get_dest_weather(
            products=[DEST_WEATHER_PRODUCT.forecast7days], query="Chicago", one_observation=True
        )

    with pytest.raises(ApiError):
        ls2 = LS(api_key="dummy")
        ls2.get_dest_weather(products=[DEST_WEATHER_PRODUCT.observation], query="Chicago")


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_ls_autosuggest():
    """Test autosuggest api."""
    ls = LS(api_key=LS_API_KEY)
    resp = ls.autosuggest(query="bar", limit=5, at=["-13.163068,-72.545128"], in_country=["USA"])
    assert resp.items
    assert len(resp.items) <= 5

    search_in_circle1 = SearchCircle(lat=52.53, lng=13.38, radius="10000")
    search_in_bbox1 = ("13.08836", "52.33812", "13.761", "52.6755")

    resp3 = ls.autosuggest(query="bar", limit=5, search_in_circle=search_in_circle1, lang=["en"])
    assert resp3.items
    assert len(resp3.items) <= 5

    resp4 = ls.autosuggest(
        query="res",
        limit=5,
        search_in_bbox=search_in_bbox1,
        terms_limit=3,
        show=[SHOW.phonemes],
        political_view=POLITICAL_VIEW.RUS,
    )
    assert resp4.items
    assert len(resp4.items) <= 5
    assert len(resp4.queryTerms) == 3

    for item in resp4.items:
        if item["resultType"] == "place":
            assert item["politicalView"]
            assert item["phonemes"]

    with pytest.raises(ValueError):
        ls.autosuggest(
            query="res",
        )

    with pytest.raises(ValueError):
        ls.autosuggest(
            query="res",
            search_in_bbox=search_in_bbox1,
            search_in_circle=search_in_circle1,
        )

    with pytest.raises(ValueError):
        ls.autosuggest(
            query="res",
            at=["-13.163068,-72.545128"],
            search_in_bbox=search_in_bbox1,
            search_in_circle=search_in_circle1,
        )

    with pytest.raises(ApiError):
        ls2 = LS(api_key="dummy")
        ls2.autosuggest(
            query="res",
            at=["-13.163068,-72.545128"],
        )


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_ls_geocoding():
    """Test geocoding api."""
    address = "200 S Mathilda Sunnyvale CA"
    ls = LS(api_key=LS_API_KEY)
    resp = ls.geocode(query=address, limit=2)
    assert isinstance(resp, GeocoderResponse)
    assert resp.__str__()
    assert resp.as_json_string()
    geo_json = resp.to_geojson()
    assert geo_json.type == "FeatureCollection"
    assert geo_json.features
    pos = resp.items[0]["position"]
    assert len(resp.items) == 1
    assert pos == {"lat": 37.37634, "lng": -122.03405}


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_ls_geocoding_exception():
    """Test geocoding api exception."""
    address = "Goregaon West, Mumbai 400062, India"

    ls = LS(api_key="dummy")
    with pytest.raises(ApiError):
        ls.geocode(query=address)
    with pytest.raises(ValueError):
        ls.geocode(query="")
        ls.geocode(query="   ")


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_ls_reverse_geocoding():
    """Test reverse geocoding."""
    ls = LS(api_key=LS_API_KEY)
    resp = ls.reverse_geocode(lat=19.1646, lng=72.8493)
    address = resp.items[0]["address"]["label"]
    assert "Goregaon" in address
    resp1 = ls.reverse_geocode(lat=19.1646, lng=72.8493, limit=4)
    assert len(resp1.items) == 4


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_ls_reverse_geocoding_exception():
    """Test reverse geocoding api exception."""
    ls = LS(api_key=LS_API_KEY)
    with pytest.raises(ValueError):
        ls.reverse_geocode(lat=91, lng=90)
    with pytest.raises(ValueError):
        ls.reverse_geocode(lat=19, lng=190)
    ls = LS(api_key="dummy")
    with pytest.raises(ApiError):
        ls.reverse_geocode(lat=19.1646, lng=72.8493)


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_isonline_routing():
    """Test isonline routing api."""
    ls = LS(api_key=LS_API_KEY)
    place_options = PlaceOptions(
        course=ROUTE_COURSE.west,
        sideof_street_hint=[52.512149, 13.304076],
        match_sideof_street=ROUTE_MATCH_SIDEOF_STREET.always,
        radius=10,
        min_course_distance=10,
    )
    assert json.loads(place_options.__str__()) == {
        "course": 270,
        "sideOfStreetHint": "52.512149,13.304076",
        "matchSideOfStreet": "always",
        "namehint": None,
        "radius": 10,
        "minCourseDistance": 10,
    }
    origin_waypoint_options = WayPointOptions(stop_duration=0)

    result = ls.calculate_isoline(
        origin=[52.5, 13.4],
        range="1000,3000",
        range_type=RANGE_TYPE.time,
        transport_mode=ISOLINE_ROUTING_TRANSPORT_MODE.car,
        departure_time=datetime.now(),
        truck=Truck(
            shipped_hazardous_goods=[SHIPPED_HAZARDOUS_GOODS.explosive],
            gross_weight=100,
            weight_per_axle=10,
            height=10,
            width=10,
            length=10,
            tunnel_category="B",
            axle_count=4,
        ),
        shape_max_points=100,
        avoid_features=[ISOLINE_ROUTING_AVOID_FEATURES.tollRoad],
        origin_place_options=place_options,
        origin_waypoint_options=origin_waypoint_options,
    )

    assert result.isolines
    assert result.departure
    coordinates = result.isolines[0]["polygons"][0]["outer"]
    assert coordinates
    geo_json = result.to_geojson()
    assert geo_json.type == "FeatureCollection"

    destination_waypoint_options = WayPointOptions(stop_duration=0)
    result2 = ls.calculate_isoline(
        destination=[52.51578, 13.37749],
        range="600",
        range_type=RANGE_TYPE.time,
        transport_mode=ISOLINE_ROUTING_TRANSPORT_MODE.car,
        destination_place_options=place_options,
        destination_waypoint_options=destination_waypoint_options,
    )
    assert result2.isolines
    assert result2.arrival

    with pytest.raises(ValueError):
        ls.calculate_isoline(
            destination=[82.8628, 135.00],
            range="3000",
            range_type=RANGE_TYPE.distance,
            transport_mode=ISOLINE_ROUTING_TRANSPORT_MODE.car,
            arrival_time=datetime.now(),
        )
    with pytest.raises(ValueError):
        ls.calculate_isoline(
            origin=[52.5, 13.4],
            range="900",
            range_type=RANGE_TYPE.time,
            transport_mode=ISOLINE_ROUTING_TRANSPORT_MODE.car,
            destination=[52.5, 13.4],
        )
    with pytest.raises(ApiError):
        ls2 = LS(api_key="dummy")
        ls2.calculate_isoline(
            origin=[52.5, 13.4],
            range="900",
            range_type=RANGE_TYPE.time,
            transport_mode=ISOLINE_ROUTING_TRANSPORT_MODE.car,
        )


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_isonline_routing_exception():
    """Test isonline exceptions."""
    ls = LS(api_key=LS_API_KEY)
    with pytest.raises(ValueError):
        ls.calculate_isoline(
            range="900",
            range_type=RANGE_TYPE.time,
            transport_mode=ISOLINE_ROUTING_TRANSPORT_MODE.car,
        )
    with pytest.raises(ValueError):
        ls.calculate_isoline(
            range="900",
            range_type=RANGE_TYPE.time,
            transport_mode=ISOLINE_ROUTING_TRANSPORT_MODE.car,
            arrival_time=datetime.now(),
            origin=[52.5, 13.4],
        )
    with pytest.raises(ValueError):
        ls.calculate_isoline(
            range="900",
            range_type=RANGE_TYPE.time,
            transport_mode=ISOLINE_ROUTING_TRANSPORT_MODE.car,
            departure_time=datetime.now(),
            destination=[52.5, 13.4],
        )


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_ls_discover():
    ls = LS(api_key=LS_API_KEY)
    result = ls.discover(query="starbucks", center=[19.1663, 72.8526], radius=10000, lang="en")
    assert len(result.items) == 20

    result2 = ls.discover(
        query="starbucks",
        center=[19.1663, 72.8526],
        country_codes=["IND"],
        limit=2,
    )
    assert len(result2.items) == 2

    result3 = ls.discover(
        query="starbucks",
        bounding_box=[13.08836, 52.33812, 13.761, 52.6755],
    )
    assert len(result3.items) == 20

    with pytest.raises(ValueError):
        ls.discover(
            query="starbucks",
            center=[52.5, 13.4],
            bounding_box=[13.08836, 52.33812, 13.761, 52.6755],
        )

    with pytest.raises(ApiError):
        ls2 = LS(api_key="dummy")
        ls2.discover(query="starbucks", center=[19.1663, 72.8526], radius=10000, limit=10)


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_ls_browse():
    ls = LS(api_key=LS_API_KEY)
    result = ls.browse(
        center=[19.1663, 72.8526],
        radius=9000,
        limit=5,
        categories=[
            PLACES_CATEGORIES.historical_monument,
            PLACES_CATEGORIES.museum,
            PLACES_CATEGORIES.park_recreation_area,
            PLACES_CATEGORIES.leisure,
            PLACES_CATEGORIES.shopping_mall,
        ],
        lang="en",
    )
    assert len(result.items) == 5

    result2 = ls.browse(
        center=[19.1663, 72.8526],
        name="starbucks",
        country_codes=["IND"],
        limit=10,
        categories=[PLACES_CATEGORIES.restaurant],
        lang="en",
    )
    assert len(result2.items) <= 10

    result3 = ls.browse(
        center=[19.1663, 72.8526],
        name="starbucks",
        bounding_box=[13.08836, 52.33812, 13.761, 52.6755],
        categories=[PLACES_CATEGORIES.restaurant],
        lang="en",
    )
    assert len(result3.items) <= 10

    with pytest.raises(ApiError):
        ls2 = LS(api_key="dummy")
        ls2.browse(
            center=[19.1663, 72.8526],
            radius=9000,
            limit=5,
            categories=[
                PLACES_CATEGORIES.historical_monument,
                PLACES_CATEGORIES.museum,
                PLACES_CATEGORIES.park_recreation_area,
                PLACES_CATEGORIES.leisure,
                PLACES_CATEGORIES.shopping_mall,
            ],
        )


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_ls_lookup():
    ls = LS(api_key=LS_API_KEY)
    result = ls.lookup(
        location_id="here:pds:place:276u0vhj-b0bace6448ae4b0fbc1d5e323998a7d2",
        lang="en",
    )
    assert result.response["title"] == "Frankfurt-Hahn Airport"

    with pytest.raises(ApiError):
        ls2 = LS(api_key="dummy")
        ls2.lookup(location_id="here:pds:place:276u0vhj-b0bace6448ae4b0fbc1d5e323998a7d2")


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_car_route():
    """Test routing API for car route."""
    ls = LS(api_key=LS_API_KEY)
    avoid_areas = [AvoidBoundingBox(68.1766451354, 7.96553477623, 97.4025614766, 35.4940095078)]
    avoid_features = [ROUTING_AVOID_FEATURES.tollRoad]
    via1 = Via(lat=52.52426, lng=13.43000)
    via2 = Via(lat=52.52624, lng=13.44012)
    result = ls.car_route(
        origin=[52.51375, 13.42462],
        destination=[52.52332, 13.42800],
        via=[via1, via2],
        return_results=[ROUTING_RETURN.polyline, ROUTING_RETURN.elevation],
        departure_time=datetime.now(),
        spans=[ROUTING_SPANS.names],
        avoid_areas=avoid_areas,
        avoid_features=avoid_features,
        exclude=["IND", "NZL", "AUS"],
    )
    assert result.response["routes"][0]["sections"][0]["departure"]["place"]["location"] == {
        "lat": 52.5137479,
        "lng": 13.4246242,
        "elv": 76.0,
    }
    assert result.response["routes"][0]["sections"][1]["departure"]["place"]["location"] == {
        "lat": 52.5242323,
        "lng": 13.4301462,
        "elv": 80.0,
    }
    assert type(result.to_geojson()) == FeatureCollection


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_car_route_extra_options():
    """Test routing API for car route."""
    place_options = PlaceOptions(
        course=ROUTE_COURSE.west,
        sideof_street_hint=[52.512149, 13.304076],
        match_sideof_street=ROUTE_MATCH_SIDEOF_STREET.always,
        radius=10,
        min_course_distance=10,
    )
    assert json.loads(place_options.__str__()) == {
        "course": 270,
        "sideOfStreetHint": "52.512149,13.304076",
        "matchSideOfStreet": "always",
        "namehint": None,
        "radius": 10,
        "minCourseDistance": 10,
    }
    via_waypoint_options = WayPointOptions(stop_duration=0, pass_through=True)
    assert json.loads(via_waypoint_options.__str__()) == {
        "stopDuration": 0,
        "passThrough": True,
    }
    dest_waypoint_options = WayPointOptions(stop_duration=10, pass_through=False)
    via1 = Via(
        lat=52.52426,
        lng=13.43000,
        place_options=place_options,
        waypoint_options=via_waypoint_options,
    )
    via2 = Via(lat=52.52426, lng=13.43000)
    via3 = Via(
        lat=52.52426,
        lng=13.43000,
        place_options=place_options,
        waypoint_options=via_waypoint_options,
    )
    ls = LS(api_key=LS_API_KEY)
    resp = ls.car_route(
        origin=[52.51375, 13.42462],
        destination=[52.52332, 13.42800],
        via=[via1, via2, via3],
        origin_place_options=place_options,
        destination_place_options=place_options,
        destination_waypoint_options=dest_waypoint_options,
        return_results=[ROUTING_RETURN.polyline, ROUTING_RETURN.elevation],
        departure_time=datetime.now(),
        spans=[ROUTING_SPANS.names],
    )
    resp = resp.response
    assert len(resp["routes"][0]["sections"]) == 2
    assert list(resp["routes"][0]["sections"][0].keys()) == [
        "id",
        "type",
        "departure",
        "arrival",
        "polyline",
        "spans",
        "notices",
        "transport",
    ]


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_bicycle_route():
    """Test routing API for car route."""
    ls = LS(api_key=LS_API_KEY)
    avoid_areas = [AvoidBoundingBox(68.1766451354, 7.96553477623, 97.4025614766, 35.4940095078)]
    avoid_features = [ROUTING_AVOID_FEATURES.tollRoad]
    via = Via(lat=52.52426, lng=13.43000)
    _ = ls.bicycle_route(
        origin=[52.51375, 13.42462],
        destination=[52.52332, 13.42800],
        via=[via],
        return_results=[ROUTING_RETURN.polyline, ROUTING_RETURN.elevation],
        departure_time=datetime.now(),
        spans=[ROUTING_SPANS.names],
        avoid_areas=avoid_areas,
        avoid_features=avoid_features,
        exclude=["IND", "NZL", "AUS"],
    )


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_truck_route():
    """Test routing API for truck route."""
    ls = LS(api_key=LS_API_KEY)
    truck = Truck(
        shipped_hazardous_goods=[SHIPPED_HAZARDOUS_GOODS.explosive],
        gross_weight=100,
        weight_per_axle=10,
        height=10,
        width=10,
        length=10,
        tunnel_category="B",
        axle_count=4,
    )
    avoid_areas = [AvoidBoundingBox(68.1766451354, 7.96553477623, 97.4025614766, 35.4940095078)]
    avoid_features = [ROUTING_AVOID_FEATURES.tollRoad]
    via = Via(lat=52.52426, lng=13.43000)
    _ = ls.truck_route(
        origin=[52.51375, 13.42462],
        destination=[52.52332, 13.42800],
        via=[via],
        return_results=[ROUTING_RETURN.polyline, ROUTING_RETURN.elevation],
        departure_time=datetime.now(),
        spans=[ROUTING_SPANS.names],
        truck=truck,
        avoid_areas=avoid_areas,
        avoid_features=avoid_features,
        exclude=["IND", "NZL", "AUS"],
    )


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_scooter_route():
    """Test routing API for scooter route."""
    ls = LS(api_key=LS_API_KEY)
    scooter = Scooter(allow_highway=True)
    assert json.loads(scooter.__str__()) == {"allowHighway": True}
    via = Via(lat=52.52426, lng=13.43000)
    _ = ls.scooter_route(
        origin=[52.51375, 13.42462],
        destination=[52.52332, 13.42800],
        via=[via],
        return_results=[ROUTING_RETURN.polyline, ROUTING_RETURN.elevation],
        departure_time=datetime.now(),
        spans=[ROUTING_SPANS.names],
        scooter=scooter,
        exclude=["IND", "NZL", "AUS"],
    )


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_pedestrian_route():
    """Test routing API for pedestrian route."""
    ls = LS(api_key=LS_API_KEY)
    via = Via(lat=52.52426, lng=13.43000)
    _ = ls.pedestrian_route(
        origin=[52.51375, 13.42462],
        destination=[52.52332, 13.42800],
        via=[via],
        return_results=[ROUTING_RETURN.polyline, ROUTING_RETURN.elevation],
        departure_time=datetime.now(),
        spans=[ROUTING_SPANS.names],
        exclude=["IND", "NZL", "AUS"],
    )


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_matrix_route_exception():
    """Test exceptions for Matrix routing."""
    ls = LS(api_key=LS_API_KEY)
    origins = [
        {"lat": 37.76, "lng": -122.42},
        {"lat": 40.63, "lng": -74.09},
        {"lat": 30.26, "lng": -97.74},
        {"lat": 40.63, "lng": -74.09},
    ]
    region_definition = CircleRegion(radius=1000, center={"lat": 37.76, "lng": -122.42})
    matrix_attributes = [MATRIX_ATTRIBUTES.distances, MATRIX_ATTRIBUTES.travelTimes]
    profile = PROFILE.carFast
    truck = Truck(
        shipped_hazardous_goods=[SHIPPED_HAZARDOUS_GOODS.explosive],
        gross_weight=100,
        weight_per_axle=10,
        height=10,
        width=10,
        length=10,
        tunnel_category="B",
        axle_count=4,
    )
    with pytest.raises(ValueError):
        ls.matrix(
            origins=origins,
            region_definition=region_definition,
            profile=profile,
            matrix_attributes=matrix_attributes,
        )

    with pytest.raises(ValueError):
        ls.matrix(
            origins=origins,
            region_definition=region_definition,
            matrix_attributes=matrix_attributes,
            transport_mode=ROUTING_TRANSPORT_MODE.car,
            truck=truck,
        )


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_matrix_route():
    """Test Matrix routing."""
    ls = LS(api_key=LS_API_KEY)
    origins = [
        {"lat": 37.76, "lng": -122.42},
        {"lat": 40.63, "lng": -74.09},
        {"lat": 30.26, "lng": -97.74},
    ]
    region_definition = WorldRegion()
    matrix_attributes = [MATRIX_ATTRIBUTES.distances, MATRIX_ATTRIBUTES.travelTimes]
    avoid_areas = AvoidBoundingBox(68.1766451354, 7.96553477623, 97.4025614766, 35.4940095078)
    assert json.loads(avoid_areas.__str__()) == {
        "type": "boundingBox",
        "north": 68.1766451354,
        "south": 7.96553477623,
        "west": 97.4025614766,
        "east": 35.4940095078,
    }
    truck = Truck(
        shipped_hazardous_goods=[SHIPPED_HAZARDOUS_GOODS.explosive],
        gross_weight=100,
        weight_per_axle=10,
        height=10,
        width=10,
        length=10,
        tunnel_category="B",
        axle_count=4,
    )
    result = ls.matrix(
        origins=origins,
        region_definition=region_definition,
        destinations=origins,
        routing_mode=ROUTING_MODE.fast,
        departure_time=datetime.now(tz=pytz.utc),
        transport_mode=ROUTING_TRANSPORT_MODE.truck,
        avoid_features=[AVOID_FEATURES.tollRoad],
        avoid_areas=[avoid_areas],
        truck=truck,
        matrix_attributes=matrix_attributes,
    )
    mat = result.matrix
    assert mat["numOrigins"] == 3
    assert mat["numDestinations"] == 3
    assert len(mat["distances"]) == 9

    profile = PROFILE.carShort
    result2 = ls.matrix(
        origins=origins,
        region_definition=region_definition,
        matrix_attributes=matrix_attributes,
        profile=profile,
    )
    mat2 = result2.matrix
    assert mat2["numOrigins"] == 3
    assert mat2["numDestinations"] == 3
    with pytest.raises(NotImplementedError):
        result2.to_geojson()
    assert isinstance(result2.to_distnaces_matrix(), pd.DataFrame)
    assert isinstance(result2.to_travel_times_matrix(), pd.DataFrame)


@pytest.mark.skipif(not LS_API_KEY, reason="No api key found.")
def test_matrix_route_async():
    """Test Matrix routing."""
    ls = LS(api_key=LS_API_KEY)
    origins = [
        {"lat": 37.76, "lng": -122.42},
        {"lat": 40.63, "lng": -74.09},
        {"lat": 30.26, "lng": -97.74},
    ]
    region_definition = WorldRegion()
    matrix_attributes = [MATRIX_ATTRIBUTES.distances, MATRIX_ATTRIBUTES.travelTimes]
    avoid_areas = AvoidBoundingBox(68.1766451354, 7.96553477623, 97.4025614766, 35.4940095078)
    truck = Truck(
        shipped_hazardous_goods=[SHIPPED_HAZARDOUS_GOODS.explosive],
        gross_weight=100,
        weight_per_axle=10,
        height=10,
        width=10,
        length=10,
        tunnel_category="B",
        axle_count=4,
    )
    result = ls.matrix(
        origins=origins,
        region_definition=region_definition,
        async_req=True,
        destinations=origins,
        routing_mode=ROUTING_MODE.fast,
        departure_time="any",
        transport_mode=ROUTING_TRANSPORT_MODE.truck,
        avoid_features=[AVOID_FEATURES.tollRoad],
        avoid_areas=[avoid_areas],
        truck=truck,
        matrix_attributes=matrix_attributes,
    )
    mat = result.matrix
    assert mat["numOrigins"] == 3
    assert mat["numDestinations"] == 3
    assert len(mat["distances"]) == 9


def test_matrix_routing_config():
    """Test Matrix routing config objects."""
    circle = CircleRegion(radius=1000, center={"lat": 37.76, "lng": -122.42})
    assert json.loads(circle.__str__()) == {
        "type": "circle",
        "center": {"lat": 37.76, "lng": -122.42},
        "radius": 1000,
    }

    bbox = BoundingBoxRegion(0, 0, 0, 0)
    assert json.loads(bbox.__str__()) == {
        "type": "boundingBox",
        "north": 0,
        "south": 0,
        "west": 0,
        "east": 0,
    }

    poly = PolygonRegion(outer=[1, 1, 1, 1, 1, 1])
    assert json.loads(poly.__str__()) == {
        "type": "polygon",
        "outer": [1, 1, 1, 1, 1, 1],
    }

    autocircle = AutoCircleRegion(margin=100)
    assert json.loads(autocircle.__str__()) == {"type": "autoCircle", "margin": 100}



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0
"""
This module implements base class for low level api client.
"""
import urllib.request
from typing import Optional, Union

import requests

from here_location_services.exceptions import AuthenticationException, TooManyRequestsException


class Api:
    """Base class for low level api calls."""

    def __init__(self, access_token, proxies: Optional[dict] = None):
        self.access_token = access_token
        self._user_agent = "dhpy"
        self.proxies: Optional[dict] = proxies or urllib.request.getproxies()

    @property
    def headers(self) -> dict:
        """
        Return HTTP request headers with Bearer token in ``Authorization``
        field.

        :return: authorization tokens
        """
        return {"Authorization": f"Bearer {self.access_token}"}

    def post(
        self,
        url: str,
        data: Optional[Union[dict, list, bytes, str]] = None,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
        **kwargs,
    ) -> requests.Response:
        """
        Perform a post request of an API at a specified URL with backoff.

        :param url: URL of the API.
        :param data: Post data for http request.
        :param params: Parameters to pass to the API.
        :param headers: Request headers. Defaults to the api headers property.
        :param kwargs: Optional arguments that request takes.
        :return: response from the API.
        """
        headers = headers or self.headers
        headers["User-Agent"] = self._user_agent
        if isinstance(data, dict) or isinstance(data, list):
            return requests.post(
                url,
                headers=headers,
                json=data,
                params=params,
                proxies=self.proxies,
                **kwargs,
            )
        else:
            return requests.post(
                url,
                headers=headers,
                data=data,
                params=params,
                proxies=self.proxies,
                **kwargs,
            )

    @staticmethod
    def raise_response_exception(resp: requests.Response) -> None:
        """
        Parse HTTP errors status code and raise necessary exceptions.

        :param resp: An HTTP response to parse.
        :raises TooManyRequestsException: If platform responds with HTTP 429.
        :raises AuthenticationException: If platform responds with HTTP 401 or 403.
        :raises Exception: If client responds with any other exception.
        """
        if resp.status_code == 429:
            raise TooManyRequestsException(resp)
        elif resp.status_code in [401, 403]:
            raise AuthenticationException(resp)
        else:
            raise Exception(resp.text)



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""
This module provides an ``Auth`` class to authenticate an app on the platform.

The authentication is based on some credentials object and will create an
access token. It can be checked if the token is still valid, and it can be
refreshed, too.
"""


from datetime import datetime, timedelta
from typing import Optional

from requests_oauthlib import OAuth1

from here_location_services.platform.apis.aaa_oauth2_api import AAAOauth2Api
from here_location_services.platform.credentials import PlatformCredentials


class Auth:
    """
    This class is responsible for authenticating with the HERE platform.

    It requires PlatformCredentials, AAAOauth2BaseApi object.
    """

    def __init__(self, credentials: PlatformCredentials, aaa_oauth2_api: AAAOauth2Api):
        """
        Instantiate authentication token.

        :param credentials: an instance of PlatformCredentials
        :param aaa_oauth2_api: an instance of AAAOauth2Api required
            in case of Credentials type.
        """
        self.credentials = credentials
        self.aaa_oauth2_api = aaa_oauth2_api

        self._token: Optional[str] = None
        self._token_type: Optional[str] = None
        self._token_expires_in: Optional[int] = None
        self._token_requested_at: Optional[datetime] = None
        self._token_expires_at: Optional[datetime] = None
        self._scope: Optional[str] = None

    @property
    def token(self) -> Optional[str]:
        """
        Return the current token or requests a new one if needed.

        :return: a valid token
        """
        if not self.token_still_valid():
            self.generate_token()
        return self._token

    def token_still_valid(self) -> bool:
        """
        Check whether the auth token is still valid or expired.

        :return: a boolean indicating if a token is still valid.
        """
        if not self._token or not self._token_expires_at:
            return False
        return datetime.now() < (self._token_expires_at - timedelta(seconds=60))

    def generate_token(self):
        """
        Authenticate with the HERE account service and retrieve a new token.
        """
        oauth = OAuth1(
            self.credentials.cred_properties["key"],
            client_secret=self.credentials.cred_properties["secret"],
            signature_method="HMAC-SHA256",
        )
        response_json = self.aaa_oauth2_api.request_scoped_access_token(
            oauth, data="grant_type=client_credentials"
        )

        self._token = response_json.get("access_token")
        self._token_type = response_json.get("token_type")
        self._token_expires_in = int(response_json.get("expires_in"))
        self._token_requested_at = datetime.now()
        self._token_expires_at = self._token_requested_at + timedelta(
            seconds=self._token_expires_in
        )



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""This module contains class to interact with Location services REST APIs."""

import os
import urllib
import urllib.request
from datetime import date, datetime
from time import sleep
from typing import Dict, List, Optional, Tuple, Union

from geojson import LineString, MultiPolygon, Point, Polygon

from here_location_services.config.routing_config import Scooter, Via
from here_location_services.platform.apis.aaa_oauth2_api import AAAOauth2Api
from here_location_services.platform.auth import Auth
from here_location_services.platform.credentials import PlatformCredentials

from .autosuggest_api import AutosuggestApi
from .config.autosuggest_config import SearchCircle
from .config.base_config import PlaceOptions, Truck, WayPointOptions
from .config.matrix_routing_config import (
    AutoCircleRegion,
    AvoidBoundingBox,
    BoundingBoxRegion,
    CircleRegion,
    PolygonRegion,
    WorldRegion,
)
from .destination_weather_api import DestinationWeatherApi
from .exceptions import ApiError
from .geocoding_search_api import GeocodingSearchApi
from .isoline_routing_api import IsolineRoutingApi
from .matrix_routing_api import MatrixRoutingApi
from .responses import (
    AutosuggestResponse,
    BrowseResponse,
    DestinationWeatherResponse,
    DiscoverResponse,
    GeocoderResponse,
    IsolineResponse,
    LookupResponse,
    MatrixRoutingResponse,
    ReverseGeocoderResponse,
    RoutingResponse,
)
from .routing_api import RoutingApi


class LS:
    """
    A single interface for the user to interact with rest of
    the Location services APIs.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        platform_credentials: Optional[PlatformCredentials] = None,
        proxies: Optional[dict] = None,
        country: str = "row",
    ):
        api_key = api_key or os.environ.get("LS_API_KEY")
        self.auth: Optional[Auth] = None
        if not api_key:
            credentials = platform_credentials or PlatformCredentials.from_default()
            aaa_oauth2_api = AAAOauth2Api(
                base_url=credentials.cred_properties["endpoint"], proxies={}
            )
            self.auth = Auth(credentials=credentials, aaa_oauth2_api=aaa_oauth2_api)

        self.proxies = proxies or urllib.request.getproxies()
        self.geo_search_api = GeocodingSearchApi(
            api_key=api_key,
            auth=self.auth,
            proxies=proxies,
            country=country,
        )
        self.isoline_routing_api = IsolineRoutingApi(
            api_key=api_key,
            auth=self.auth,
            proxies=proxies,
            country=country,
        )
        self.routing_api = RoutingApi(
            api_key=api_key,
            auth=self.auth,
            proxies=proxies,
            country=country,
        )
        self.matrix_routing_api = MatrixRoutingApi(
            api_key=api_key, auth=self.auth, proxies=proxies, country=country
        )
        self.autosuggest_api = AutosuggestApi(
            api_key=api_key,
            auth=self.auth,
            proxies=proxies,
            country=country,
        )
        self.destination_weather_api = DestinationWeatherApi(
            api_key=api_key,
            auth=self.auth,
            proxies=proxies,
            country=country,
        )

    def geocode(self, query: str, limit: int = 20, lang: str = "en-US") -> GeocoderResponse:
        """Calculate coordinates as result of geocoding for the given ``query``.

        :param query: A string containing the input query.
        :param limit: An int representing maximum number of results to be returned.
            Default value is 20.
        :param lang: A string to represent language to be used for result rendering from
            a list of BCP47 compliant Language Codes.
        :raises ValueError: If ``query`` is empty or having all whitespace characters.
        :return: :class:`GeocoderResponse` object.
        """
        if not query or query.isspace():
            raise ValueError(f"Invalid input query: {query}")

        resp = self.geo_search_api.get_geocoding(query, limit=limit, lang=lang)
        return GeocoderResponse.new(resp.json())

    def reverse_geocode(
        self, lat: float, lng: float, limit: int = 1, lang: str = "en-US"
    ) -> ReverseGeocoderResponse:
        """
        Return the address label string as the result of reverse-geocoding the
        given ``latitude`` and ``longitude``.

        :param lat: A float representing latitude of point.
        :param lng: A float representing longitude of point.
        :param limit: An int representing maximum number of results to be returned.
            Default value is 1.
        :param lang: A string to represent language to be used for result rendering from
            a list of BCP47 compliant Language Codes.
        :raises ValueError: If Latitude is not in range between -90 and 90 or
             Longitude is not in range between -180 and 180.
        :return: :class:`ReverseGeocoderResponse` object.
        """
        if not -90 <= lat <= 90:
            raise ValueError("Latitude must be in range -90 to 90.")
        if not -180 <= lng <= 180:
            raise ValueError("Longitude must be in range -180 to 180.")

        resp = self.geo_search_api.get_reverse_geocoding(lat=lat, lng=lng, limit=limit, lang=lang)
        return ReverseGeocoderResponse.new(resp.json())

    def calculate_isoline(
        self,
        range: str,
        range_type: str,
        transport_mode: str,
        origin: Optional[List] = None,
        departure_time: Optional[datetime] = None,
        destination: Optional[List] = None,
        arrival_time: Optional[datetime] = None,
        routing_mode: Optional[str] = "fast",
        shape_max_points: Optional[int] = None,
        optimised_for: Optional[str] = "balanced",
        avoid_features: Optional[List[str]] = None,
        truck: Optional[Truck] = None,
        origin_place_options: Optional[PlaceOptions] = None,
        origin_waypoint_options: Optional[WayPointOptions] = None,
        destination_place_options: Optional[PlaceOptions] = None,
        destination_waypoint_options: Optional[WayPointOptions] = None,
    ) -> IsolineResponse:
        """Calculate isoline routing.

        Request a polyline that connects the endpoints of all routes
        leaving from one defined center with either a specified length
        or specified travel time.

        :param range: A string representing a range of isoline, unit is defined by
            parameter range type. Example: range='1000' or range='1000,2000,3000'
        :param range_type: A string representing a type of ``range``. Possible values are
            ``distance``, ``time`` and ``consumption``. For distance the unit meters. For a
            time the unit is seconds. For consumption, it is defined by the consumption
            model.
        :param transport_mode: A string representing Mode of transport to be used for the
            calculation of the isolines.
            Example: ``car``.
        :param origin: Center of the isoline request. The Isoline(s) will cover the region
            which can be reached from this point within given range. It cannot be used in
            combination with ``destination`` parameter.
        :param departure_time: Specifies the time of departure as defined by either date-time
            or full-date partial-time in RFC 3339, section 5.6 (for example, 2019-06-24T01:23:45).
            The requested time is converted to the local time at origin. When the optional timezone
            offset is not specified, time is assumed to be local. If neither departure_time or
            arrival_time are specified, current time at departure location will be used. All Time
            values in the response are returned in the timezone of each location.
        :param destination: Center of the isoline request. The Isoline(s) will cover the
            region within the specified range that can reach this point. It cannot be used
            in combination with ``origin`` parameter.
        :param arrival_time: Specifies the time of arrival as defined by either date-time or
            full-date T partial-time in RFC 3339, section 5.6 (for example, 2019-06-24T01:23:45).
            The requested time is converted to the local time at destination. When the optional
            timezone offset is not specified, time is assumed to be local. All Time values in
            the response are returned in the timezone of each location.
        :param routing_mode: A string to represent routing mode.
        :param shape_max_points: An integer to Limit the number of points in the resulting isoline
            geometry. If the isoline consists of multiple components, the sum of points from all
            components is considered. This parameter doesn't affect performance.
        :param optimised_for: A string to specify how isoline calculation is optimized.
        :param avoid_features: Avoid routes that violate these properties. Avoid features
            are defined in :attr:
            `AVOID_FEATURES <here_location_services.config.isoline_routing_config.AVOID_FEATURES>`
        :param truck: Different truck options to use during route calculation when transport_mode
            = truck. use object of :class:`Truck here_location_services.config.base_config.Truck>`
        :param origin_place_options: :class:`PlaceOptions` optinal place options for ``origin``.
        :param origin_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``origin``.
        :param destination_place_options: :class:`PlaceOptions` optinal place options
            for ``destination``.
        :param destination_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``destination``.
        :raises ValueError: If ``origin`` and ``destination`` are provided together.
        :return: :class:`IsolineResponse` object.
        """

        if origin and destination:
            raise ValueError("`origin` and `destination` can not be provided together.")
        if origin is None and destination is None:
            raise ValueError("please provide either `origin` or `destination`.")
        if departure_time and origin is None:
            raise ValueError("`departure_time` must be provided with `origin`")
        if arrival_time and destination is None:
            raise ValueError("`arrival` must be provided with `destination`")

        resp = self.isoline_routing_api.get_isoline_routing(
            range=range,
            range_type=range_type,
            transport_mode=transport_mode,
            origin=origin,
            departure_time=departure_time,
            destination=destination,
            arrival_time=arrival_time,
            routing_mode=routing_mode,
            shape_max_points=shape_max_points,
            optimised_for=optimised_for,
            avoid_features=avoid_features,
            truck=truck,
            origin_place_options=origin_place_options,
            origin_waypoint_options=origin_waypoint_options,
            destination_place_options=destination_place_options,
            destination_waypoint_options=destination_waypoint_options,
        )
        response = IsolineResponse.new(resp.json())

        if response.notices:
            raise ValueError("Isolines could not be calculated.")
        return response

    def autosuggest(
        self,
        query: str,
        at: Optional[List] = None,
        search_in_circle: Optional[SearchCircle] = None,
        search_in_bbox: Optional[Tuple] = None,
        in_country: Optional[List[str]] = None,
        limit: Optional[int] = 20,
        terms_limit: Optional[int] = None,
        lang: Optional[List[str]] = None,
        political_view: Optional[str] = None,
        show: Optional[List[str]] = None,
    ) -> AutosuggestResponse:
        """Suggest address or place candidates based on an incomplete or misspelled query

        :param query: A string for free-text query. Example: res, rest
        :param at: Specify the center of the search context expressed as list of coordinates
            One of `at`, `search_in_circle` or `search_in_bbox` is required.
            Parameters "at", "search_in_circle" and "search_in_bbox" are mutually exclusive. Only
            one of them is allowed.
        :param search_in_circle: Search within a circular geographic area provided as
            latitude, longitude, and radius (in meters)
        :param search_in_bbox: Search within a rectangular bounding box geographic area provided
            as tuple of west longitude, south latitude, east longitude, north latitude
        :param in_country: Search within a specific or multiple countries provided as
            comma-separated ISO 3166-1 alpha-3 country codes. The country codes are to be
            provided in all uppercase. Must be accompanied by exactly one of
            `at`, `search_in_circle` or `search_in_bbox`.
        :param limit: An integer specifiying maximum number of results to be returned.
        :param terms_limit: An integer specifiying maximum number of Query Terms Suggestions
            to be returned.
        :param lang: List of strings to select the language to be used for result rendering
            from a list of BCP 47 compliant language codes.
        :param political_view: Toggle the political view.
        :param show: Select additional fields to be rendered in the response. Please note
            that some of the fields involve additional webservice calls and can increase
            the overall response time.
        :return: :class:`requests.Response` object.
        :raises ValueError: If ``search_in_circle``,``search_in_bbox`` and ``destination``
            are provided together.
        """

        i = iter([search_in_circle, search_in_bbox, at])
        if not (any(i) and not any(i)):
            raise ValueError(
                "Exactly one of `search_in_circle` or `search_in_bbox` or `at` must be provided."
            )

        resp = self.autosuggest_api.get_autosuggest(
            query=query,
            at=at,
            search_in_bbox=search_in_bbox,
            search_in_circle=search_in_circle,
            in_country=in_country,
            limit=limit,
            terms_limit=terms_limit,
            lang=lang,
            political_view=political_view,
            show=show,
        )
        response = AutosuggestResponse.new(resp.json())

        return response

    def get_dest_weather(
        self,
        products: List[str],
        at: Optional[List] = None,
        query: Optional[str] = None,
        zipcode: Optional[str] = None,
        hourly_date: Optional[Union[date, datetime]] = None,
        one_observation: Optional[bool] = None,
        language: Optional[str] = None,
        units: Optional[str] = None,
    ) -> DestinationWeatherResponse:
        """Retrieves weather reports, weather forecasts, severe weather alerts
            and moon and sun rise and set information.

        :param products: List of :class:`DestWeatherProduct` identifying the type of
            report to obtain.
        :param at: A list of ``latitude`` and ``longitude`` specifying the area covered
            by the weather report.
        :param query: Free text query. Examples: "125, Berliner, berlin", "Beacon, Boston"
        :param zipcode: ZIP code of the location. This parameter is supported only for locations in
            the United States of America.
        :param hourly_date: Date for which hourly forecasts are to be retrieved. Can be either a
            `date` or `datetime` object
        :param one_observation: Boolean, if set to true, the response only includes the closest
            location. Only available when the `product` parameter is set to
            `DEST_WEATHER_PRODUCT.observation`.
        :param language: Defines the language used in the descriptions in the response.
        :param units: Defines whether units or imperial units are used in the response.
        :raises ValueError: If neither `at`, `query` or `zipcode` are passed.
        :raises ValueError: If `one_observation` is set to true without passing
            DEST_WEATHER_PRODUCT.observation in `products`
        :return: :class:`DestinationWeatherResponse` object.
        """

        if at is None and query is None and zipcode is None:
            raise ValueError("please provide either `at` or `query` or `zipcode`.")
        if "observation" not in products and one_observation:
            raise ValueError(
                "`one_observation` can only be set when the `products` parameter "
                + "is set to DEST_WEATHER_PRODUCT.observation"
            )

        resp = self.destination_weather_api.get_dest_weather(
            products=products,
            at=at,
            query=query,
            zipcode=zipcode,
            hourly_date=hourly_date,
            one_observation=one_observation,
            language=language,
            units=units,
        )
        response = DestinationWeatherResponse.new(resp.json())
        return response

    def get_weather_alerts(
        self,
        geometry: Union[Point, LineString, Polygon, MultiPolygon],
        start_time: datetime,
        id: Optional[str] = None,
        weather_severity: Optional[int] = None,
        weather_type: Optional[str] = None,
        country: Optional[str] = None,
        end_time: Optional[datetime] = None,
        width: Optional[int] = 50000,
    ) -> DestinationWeatherResponse:
        """Retrieves weather reports, weather forecasts, severe weather alerts
            and moon and sun rise and set information.

        :param geometry: Point or LineString or Polygon or MultiPolygon defining the route or
            a single location
        :param start_time: Start time of the event
        :param id: Unique weather alert id.
        :param weather_severity: Defines the severity of the weather event as defined
            in :class:`WeatherSeverity`.
        :param weather_type: Defines the type of the weather event as defined
            in :class:`WeatherType`.
        :param country: String for ISO-3166-1 2-letter country code.
        :param end_time: End time of the event. If not present, warning is valid until
            it is not removed from the feed by national weather institutes
            (valid until warning is present in the response)
        :param width: int. default 50000
        :return: :class:`DestinationWeatherResponse` object.
        """

        resp = self.destination_weather_api.get_weather_alerts(
            geometry=geometry,
            id=id,
            weather_severity=weather_severity,
            weather_type=weather_type,
            country=country,
            start_time=start_time,
            end_time=end_time,
            width=width,
        )
        response = DestinationWeatherResponse.new(resp.json())
        return response

    def discover(
        self,
        query: str,
        center: Optional[List[float]] = None,
        radius: Optional[int] = None,
        country_codes: Optional[List] = None,
        bounding_box: Optional[List[float]] = None,
        limit: Optional[int] = None,
        lang: Optional[str] = None,
    ) -> DiscoverResponse:
        """Search places using Location Services discover endpoint.

        This method uses location services ``discover`` endpoint to search places based on
        query which is free-form text.
        There are three different combination of inputs as shown below to search places
        using discover:

        - ``center`` and ``country_code``
        - ``center`` and ``radius``
        - ``bounding_box``

        :param query: A string representing free-text query to search places.
        :param center: A list of latitude and longitude representing the center for
            search query.
        :param radius: A radius in meters along with center for searching places.
        :param country_codes: A list of  ISO 3166-1 alpha-3 country codes.
        :param bounding_box: A bounding box, provided as west longitude, south latitude,
            east longitude, north latitude.
        :param limit: An int representing maximum number of results to be returned.
        :param lang: A string to represent language to be used for result rendering from
            a list of BCP47 compliant Language Codes.
        :raises ValueError: If ``center`` and ``bounding_box`` are provided together.
        :return: :class:`DiscoverResponse` object.
        """
        if center and bounding_box:
            raise ValueError(
                f"Params: center:{center} and bounding_box:{bounding_box} "
                f"can not be provided together."
            )

        resp = self.geo_search_api.get_search_discover(
            query=query,
            center=center,
            radius=radius,
            country_codes=country_codes,
            bounding_box=bounding_box,
            limit=limit,
            lang=lang,
        )
        return DiscoverResponse.new(resp.json())

    def browse(
        self,
        center: List,
        radius: Optional[int] = None,
        country_codes: Optional[List] = None,
        bounding_box: Optional[List[float]] = None,
        categories: Optional[List] = None,
        limit: Optional[int] = None,
        name: Optional[str] = None,
        lang: Optional[str] = None,
    ) -> BrowseResponse:
        """Get search results for places based on different filters such as categories or name.

        :param center: A list of latitude and longitude representing the center for
            search query.
        :param radius: A radius in meters along with center for searching places.
        :param country_codes: A list of  ISO 3166-1 alpha-3 country codes.
        :param bounding_box: A bounding box, provided as west longitude, south latitude,
            east longitude, north latitude.
        :param categories: A list strings of category-ids.
        :param limit: An int representing maximum number of results to be returned.
        :param name: A string representing Full-text filter on POI names/titles.
        :param lang: A string to represent language to be used for result rendering from
            a list of BCP47 compliant Language Codes.
        :return: :class:`BrowseResponse` object.
        """
        resp = self.geo_search_api.get_search_browse(
            center=center,
            radius=radius,
            country_codes=country_codes,
            bounding_box=bounding_box,
            categories=categories,
            limit=limit,
            name=name,
            lang=lang,
        )
        return BrowseResponse.new(resp.json())

    def lookup(self, location_id: str, lang: Optional[str] = None) -> LookupResponse:
        """
        Get search results by providing ``location_id``.

        :param location_id: A string representing id.
        :param lang: A string to represent language to be used for result rendering from
            a list of BCP47 compliant Language Codes.
        :return: :class:`LookupResponse` object.
        """
        resp = self.geo_search_api.get_search_lookup(location_id=location_id, lang=lang)
        return LookupResponse.new(resp.json())

    def car_route(
        self,
        origin: List,
        destination: List,
        via: Optional[List[Via]] = None,
        origin_place_options: Optional[PlaceOptions] = None,
        origin_waypoint_options: Optional[WayPointOptions] = None,
        destination_place_options: Optional[PlaceOptions] = None,
        destination_waypoint_options: Optional[WayPointOptions] = None,
        departure_time: Optional[datetime] = None,
        routing_mode: str = "fast",
        alternatives: int = 0,
        units: str = "metric",
        lang: str = "en-US",
        return_results: Optional[List] = None,
        spans: Optional[List] = None,
        avoid_features: Optional[List[str]] = None,
        avoid_areas: Optional[List[AvoidBoundingBox]] = None,
        exclude: Optional[List[str]] = None,
    ) -> RoutingResponse:
        """Calculate ``car`` route between two endpoints.

        :param origin: A list of ``latitude`` and ``longitude`` of origin point of route.
        :param destination: A list of ``latitude`` and ``longitude`` of destination point of route.
        :param via: A list of :class:`Via` objects.
        :param origin_place_options: :class:`PlaceOptions` optinal place options for ``origin``.
        :param origin_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``origin``.
        :param destination_place_options: :class:`PlaceOptions` optinal place options
            for ``destination``.
        :param destination_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``destination``.
        :param departure_time: :class:`datetime.datetime` object.
        :param routing_mode: A string to represent routing mode. use config defined in :attr:`ROUTING_MODE <here_location_services.config.routing_config.ROUTING_MODE>`
        :param alternatives: Number of alternative routes to return aside from the optimal route.
            default value is ``0`` and maximum is ``6``.
        :param units: A string representing units of measurement used in guidance instructions.
            The default is metric.
        :param lang: A string representing preferred language of the response.
            The value should comply with the IETF BCP 47.
        :param return_results: A list of strings.
        :param spans: A list of strings to define which attributes are included in the response
            spans. use config defined in :attr:`ROUTING_SPANS <here_location_services.config.routing_config.ROUTING_SPANS>`
        :param avoid_features: Avoid routes that violate these properties. Avoid features are
            defined in :attr:`AVOID_FEATURES <here_location_services.config.routing_config.AVOID_FEATURES>`
        :param avoid_areas: A list of areas to avoid during route calculation. To define avoid area.
        :param exclude: A comma separated list of three-letter country codes
            (ISO-3166-1 alpha-3 code) that routes will exclude.
        :return: :class:`RoutingResponse` object.
        """  # noqa: E501

        resp = self.routing_api.route(
            transport_mode="car",
            origin=origin,
            destination=destination,
            via=via,
            origin_place_options=origin_place_options,
            origin_waypoint_options=origin_waypoint_options,
            destination_place_options=destination_place_options,
            destination_waypoint_options=destination_waypoint_options,
            departure_time=departure_time,
            routing_mode=routing_mode,
            alternatives=alternatives,
            units=units,
            lang=lang,
            return_results=return_results,
            spans=spans,
            avoid_features=avoid_features,
            avoid_areas=avoid_areas,
            exclude=exclude,
        )
        return RoutingResponse.new(resp.json())

    def bicycle_route(
        self,
        origin: List,
        destination: List,
        via: Optional[List[Via]] = None,
        origin_place_options: Optional[PlaceOptions] = None,
        origin_waypoint_options: Optional[WayPointOptions] = None,
        destination_place_options: Optional[PlaceOptions] = None,
        destination_waypoint_options: Optional[WayPointOptions] = None,
        departure_time: Optional[datetime] = None,
        routing_mode: str = "fast",
        alternatives: int = 0,
        units: str = "metric",
        lang: str = "en-US",
        return_results: Optional[List] = None,
        spans: Optional[List] = None,
        avoid_features: Optional[List[str]] = None,
        avoid_areas: Optional[List[AvoidBoundingBox]] = None,
        exclude: Optional[List[str]] = None,
    ) -> RoutingResponse:
        """Calculate ``bicycle`` route between two endpoints.

        :param origin: A list of ``latitude`` and ``longitude`` of origin point of route.
        :param destination: A list of ``latitude`` and ``longitude`` of destination point of route.
        :param via: A list of :class:`Via` objects.
        :param origin_place_options: :class:`PlaceOptions` optinal place options for ``origin``.
        :param origin_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``origin``.
        :param destination_place_options: :class:`PlaceOptions` optinal place options
            for ``destination``.
        :param destination_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``destination``.
        :param departure_time: :class:`datetime.datetime` object.
        :param routing_mode: A string to represent routing mode.
        :param alternatives: Number of alternative routes to return aside from the optimal route.
            default value is ``0`` and maximum is ``6``.
        :param units: A string representing units of measurement used in guidance instructions.
            The default is metric.
        :param lang: A string representing preferred language of the response.
            The value should comply with the IETF BCP 47.
        :param return_results: A list of strings.
        :param spans: A list of strings to define which attributes are included in the response
            spans.
        :param avoid_features: Avoid routes that violate these properties. Avoid features are
            defined in :attr:`AVOID_FEATURES <here_location_services.config.routing_config.AVOID_FEATURES>`
        :param avoid_areas: A list of areas to avoid during route calculation. To define avoid area.
        :param exclude: A comma separated list of three-letter country codes
            (ISO-3166-1 alpha-3 code) that routes will exclude.
        :return: :class:`RoutingResponse` object.
        """  # noqa E501
        resp = self.routing_api.route(
            transport_mode="bicycle",
            origin=origin,
            destination=destination,
            via=via,
            origin_place_options=origin_place_options,
            origin_waypoint_options=origin_waypoint_options,
            destination_place_options=destination_place_options,
            destination_waypoint_options=destination_waypoint_options,
            departure_time=departure_time,
            routing_mode=routing_mode,
            alternatives=alternatives,
            units=units,
            lang=lang,
            return_results=return_results,
            spans=spans,
            avoid_features=avoid_features,
            avoid_areas=avoid_areas,
            exclude=exclude,
        )
        return RoutingResponse.new(resp.json())

    def truck_route(
        self,
        origin: List,
        destination: List,
        via: Optional[List[Via]] = None,
        origin_place_options: Optional[PlaceOptions] = None,
        origin_waypoint_options: Optional[WayPointOptions] = None,
        destination_place_options: Optional[PlaceOptions] = None,
        destination_waypoint_options: Optional[WayPointOptions] = None,
        departure_time: Optional[datetime] = None,
        routing_mode: str = "fast",
        alternatives: int = 0,
        units: str = "metric",
        lang: str = "en-US",
        return_results: Optional[List] = None,
        spans: Optional[List] = None,
        truck: Optional[Truck] = None,
        avoid_features: Optional[List[str]] = None,
        avoid_areas: Optional[List[AvoidBoundingBox]] = None,
        exclude: Optional[List[str]] = None,
    ) -> RoutingResponse:
        """Calculate ``truck`` route between two endpoints.

        :param origin: A list of ``latitude`` and ``longitude`` of origin point of route.
        :param destination: A list of ``latitude`` and ``longitude`` of destination point of route.
        :param via: A list of :class:`Via` objects.
        :param origin_place_options: :class:`PlaceOptions` optinal place options for ``origin``.
        :param origin_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``origin``.
        :param destination_place_options: :class:`PlaceOptions` optinal place options
            for ``destination``.
        :param destination_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``destination``.
        :param departure_time: :class:`datetime.datetime` object.
        :param routing_mode: A string to represent routing mode.
        :param alternatives: Number of alternative routes to return aside from the optimal route.
            default value is ``0`` and maximum is ``6``.
        :param units: A string representing units of measurement used in guidance instructions.
            The default is metric.
        :param lang: A string representing preferred language of the response.
            The value should comply with the IETF BCP 47.
        :param return_results: A list of strings.
        :param spans: A list of strings to define which attributes are included in the response
            spans.
        :param truck: Different truck options to use during route calculation.
            use object of :class:`Truck here_location_services.config.matrix_routing_config.Truck>`
        :param avoid_features: Avoid routes that violate these properties. Avoid features are
            defined in :attr:`AVOID_FEATURES <here_location_services.config.routing_config.AVOID_FEATURES>`
        :param avoid_areas: A list of areas to avoid during route calculation. To define avoid area.
        :param exclude: A comma separated list of three-letter country codes
            (ISO-3166-1 alpha-3 code) that routes will exclude.
        :return: :class:`RoutingResponse` object.
        """  # noqa E501
        resp = self.routing_api.route(
            transport_mode="truck",
            origin=origin,
            destination=destination,
            via=via,
            origin_place_options=origin_place_options,
            origin_waypoint_options=origin_waypoint_options,
            destination_place_options=destination_place_options,
            destination_waypoint_options=destination_waypoint_options,
            departure_time=departure_time,
            routing_mode=routing_mode,
            alternatives=alternatives,
            units=units,
            lang=lang,
            return_results=return_results,
            spans=spans,
            truck=truck,
            avoid_features=avoid_features,
            avoid_areas=avoid_areas,
            exclude=exclude,
        )
        return RoutingResponse.new(resp.json())

    def scooter_route(
        self,
        origin: List,
        destination: List,
        via: Optional[List[Via]] = None,
        origin_place_options: Optional[PlaceOptions] = None,
        origin_waypoint_options: Optional[WayPointOptions] = None,
        destination_place_options: Optional[PlaceOptions] = None,
        destination_waypoint_options: Optional[WayPointOptions] = None,
        scooter: Optional[Scooter] = None,
        departure_time: Optional[datetime] = None,
        routing_mode: str = "fast",
        alternatives: int = 0,
        units: str = "metric",
        lang: str = "en-US",
        return_results: Optional[List] = None,
        spans: Optional[List] = None,
        avoid_features: Optional[List[str]] = None,
        avoid_areas: Optional[List[AvoidBoundingBox]] = None,
        exclude: Optional[List[str]] = None,
    ) -> RoutingResponse:
        """Calculate ``scooter`` route between two endpoints.

        :param origin: A list of ``latitude`` and ``longitude`` of origin point of route.
        :param destination: A list of ``latitude`` and ``longitude`` of destination point of route.
        :param via: A list of :class:`Via` objects.
        :param origin_place_options: :class:`PlaceOptions` optinal place options for ``origin``.
        :param origin_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``origin``.
        :param destination_place_options: :class:`PlaceOptions` optinal place options
            for ``destination``.
        :param destination_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``destination``.
        :param scooter: Additional attributes for scooter route.
        :param departure_time: :class:`datetime.datetime` object.
        :param routing_mode: A string to represent routing mode.
        :param alternatives: Number of alternative routes to return aside from the optimal route.
            default value is ``0`` and maximum is ``6``.
        :param units: A string representing units of measurement used in guidance instructions.
            The default is metric.
        :param lang: A string representing preferred language of the response.
            The value should comply with the IETF BCP 47.
        :param return_results: A list of strings.
        :param spans: A list of strings to define which attributes are included in the response
            spans.
        :param avoid_features: Avoid routes that violate these properties. Avoid features are
            defined in :attr:`AVOID_FEATURES <here_location_services.config.routing_config.AVOID_FEATURES>`
        :param avoid_areas: A list of areas to avoid during route calculation. To define avoid area.
        :param exclude: A comma separated list of three-letter country codes
            (ISO-3166-1 alpha-3 code) that routes will exclude.
        :return: :class:`RoutingResponse` object.
        """  # noqa E501
        resp = self.routing_api.route(
            transport_mode="scooter",
            origin=origin,
            destination=destination,
            via=via,
            origin_place_options=origin_place_options,
            origin_waypoint_options=origin_waypoint_options,
            destination_place_options=destination_place_options,
            destination_waypoint_options=destination_waypoint_options,
            scooter=scooter,
            departure_time=departure_time,
            routing_mode=routing_mode,
            alternatives=alternatives,
            units=units,
            lang=lang,
            return_results=return_results,
            spans=spans,
            avoid_features=avoid_features,
            avoid_areas=avoid_areas,
            exclude=exclude,
        )
        return RoutingResponse.new(resp.json())

    def pedestrian_route(
        self,
        origin: List,
        destination: List,
        via: Optional[List[Via]] = None,
        origin_place_options: Optional[PlaceOptions] = None,
        origin_waypoint_options: Optional[WayPointOptions] = None,
        destination_place_options: Optional[PlaceOptions] = None,
        destination_waypoint_options: Optional[WayPointOptions] = None,
        departure_time: Optional[datetime] = None,
        routing_mode: str = "fast",
        alternatives: int = 0,
        units: str = "metric",
        lang: str = "en-US",
        return_results: Optional[List] = None,
        spans: Optional[List] = None,
        avoid_features: Optional[List[str]] = None,
        avoid_areas: Optional[List[AvoidBoundingBox]] = None,
        exclude: Optional[List[str]] = None,
    ) -> RoutingResponse:
        """Calculate ``pedestrian`` route between two endpoints.

        :param origin: A list of ``latitude`` and ``longitude`` of origin point of route.
        :param destination: A list of ``latitude`` and ``longitude`` of destination point of route.
        :param via: A list of :class:`Via` objects.
        :param origin_place_options: :class:`PlaceOptions` optinal place options for ``origin``.
        :param origin_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``origin``.
        :param destination_place_options: :class:`PlaceOptions` optinal place options
            for ``destination``.
        :param destination_waypoint_options: :class:`WayPointOptions` optional waypoint options
            for ``destination``.
        :param departure_time: :class:`datetime.datetime` object.
        :param routing_mode: A string to represent routing mode.
        :param alternatives: Number of alternative routes to return aside from the optimal route.
            default value is ``0`` and maximum is ``6``.
        :param units: A string representing units of measurement used in guidance instructions.
            The default is metric.
        :param lang: A string representing preferred language of the response.
            The value should comply with the IETF BCP 47.
        :param return_results: A list of strings.
        :param spans: A list of strings to define which attributes are included in the response
            spans.
        :param avoid_features: Avoid routes that violate these properties. Avoid features are
            defined in :attr:`AVOID_FEATURES <here_location_services.config.routing_config.AVOID_FEATURES>`
        :param avoid_areas: A list of areas to avoid during route calculation. To define avoid area.
        :param exclude: A comma separated list of three-letter country codes
            (ISO-3166-1 alpha-3 code) that routes will exclude.
        :return: :class:`RoutingResponse` object.
        """  # noqa E501
        resp = self.routing_api.route(
            transport_mode="pedestrian",
            origin=origin,
            destination=destination,
            via=via,
            origin_place_options=origin_place_options,
            origin_waypoint_options=origin_waypoint_options,
            destination_place_options=destination_place_options,
            destination_waypoint_options=destination_waypoint_options,
            departure_time=departure_time,
            routing_mode=routing_mode,
            alternatives=alternatives,
            units=units,
            lang=lang,
            return_results=return_results,
            spans=spans,
            avoid_features=avoid_features,
            avoid_areas=avoid_areas,
            exclude=exclude,
        )
        return RoutingResponse.new(resp.json())

    def matrix(
        self,
        origins: List[Dict],
        region_definition: Union[
            CircleRegion,
            BoundingBoxRegion,
            PolygonRegion,
            AutoCircleRegion,
            WorldRegion,
        ],
        async_req: bool = False,
        destinations: Optional[List[Dict]] = None,
        profile: Optional[str] = None,
        departure_time: Optional[Union[datetime, str]] = None,
        routing_mode: Optional[str] = None,
        transport_mode: Optional[str] = None,
        avoid_features: Optional[List[str]] = None,
        avoid_areas: Optional[List[AvoidBoundingBox]] = None,
        truck: Optional[Truck] = None,
        matrix_attributes: Optional[List[str]] = None,
    ) -> MatrixRoutingResponse:
        """
        Calculate routing matrix between multiple ``origins`` and ``destinations`` using
        synchronous and asynchronous requests.

        A routing matrix is a matrix with rows labeled by origins and columns by destinations.
        Each entry of the matrix is travel time or distance from the origin to the destination.
        The response contains 2 optional flat arrays ``TravelTimes`` and ``distances`` depending
        upon the specified ``matrix_attributes``. Each array represents a 2D matrix where rows (i)
        corresponds to ``origins`` and columns (j) to destinations. The kth position in the array
        corresponds to the (i, j) position in the matrix defined by the following relationship:
        k = num_destitions * i + j.

        :param origins: A list of dictionaries containing lat and long for origin points.
        :param region_definition: Definition of a region in which the matrix will be calculated.
            Use object of atleast one of the following regions:

            :class:`CircleRegion <here_location_services.config.matrix_routing_config.CircleRegion>`

            :class:`BoundingBoxRegion <here_location_services.config.matrix_routing_config.BoundingBoxRegion>`

            :class:`PolygonRegion <here_location_services.config.matrix_routing_config.PolygonRegion>`

            :class:`AutoCircleRegion <here_location_services.config.matrix_routing_config.AutoCircleRegion>`

            :class:`WorldRegion <here_location_services.config.matrix_routing_config.WorldRegion>`
        :param async_req: If set to True reuqests will be sent to asynchronous matrix routing API
            else It will be sent to synchronous matrix routing API. For larger matrices, or longer
            routes, or routes in denser road networks, it is recommended to set to True.
        :param destinations: A list of dictionaries containing lat and long for destination points.
            When no destinations are specified the matrix is assumed to be quadratic with origins
            used as destinations.
        :param profile: A string to represent profile id. A set predefined profile ids for route
            calculation can be used from config
            :attr:`PROFILE <here_location_services.config.matrix_routing_config.PROFILE>`
        :param departure_time: :class:`datetime.datetime` object with explicit timezone. When
            departure_time is not specified, it is implicitly assumed to be the current time.
            The special value ``any`` enforces non time-aware routing.
        :param routing_mode: A string to represent routing mode. Routing mode values are defined
            in :attr:`ROUTING_MODE <here_location_services.config.routing_config.ROUTING_MODE>`
        :param transport_mode: A string to represent transport mode. Transport modes are defined
            in :attr:`ROUTING_TRANSPORT_MODE <here_location_services.config.routing_config.ROUTING_TRANSPORT_MODE>`
        :param avoid_features: Avoid routes that violate these properties. Avoid features are
            defined in :attr:`AVOID_FEATURES <here_location_services.config.matrix_routing_config.AVOID_FEATURES>`
        :param avoid_areas: A list of areas to avoid during route calculation. To define avoid area
            use object of :class:`AvoidBoundingBox <here_location_services.config.matrix_routing_config.AvoidBoundingBox>`
        :param truck: Different truck options to use during route calculation when
            transport_mode = truck. use object of :class:`Truck <here_location_services.config.matrix_routing_config.Truck>`
        :param matrix_attributes: Defines which attributes are included in the response as part of
            the data representation of the matrix entries summaries. Matrix attributes are defined
            in :attr:`MATRIX_ATTRIBUTES <here_location_services.config.matrix_routing_config.MATRIX_ATTRIBUTES>`
        :raises ValueError: If conflicting options are provided.
        :raises ApiError: If API response status code is not as expected.
        :return: :class:`MatrixRoutingResponse` object.
        """  # noqa E501
        if profile and type(region_definition) != WorldRegion:
            raise ValueError("profile must be used with WorldRegion only.")
        if truck and transport_mode != "truck":
            raise ValueError("Truck option must be used when transport_mode is truck")
        if async_req is True:
            resp = self.matrix_routing_api.matrix_route_async(
                origins=origins,
                region_definition=region_definition,
                destinations=destinations,
                profile=profile,
                departure_time=departure_time,
                routing_mode=routing_mode,
                transport_mode=transport_mode,
                avoid_features=avoid_features,
                avoid_areas=avoid_areas,
                truck=truck,
                matrix_attributes=matrix_attributes,
            )
            status_url = resp["statusUrl"]
            while True:
                resp_status = self.matrix_routing_api.get_async_matrix_route_status(status_url)
                if resp_status.status_code == 200 and resp_status.json().get("error"):
                    raise ApiError(resp_status)
                elif resp_status.status_code == 303:
                    result_url = resp_status.json()["resultUrl"]
                    break
                elif resp_status.status_code in (401, 403, 404, 500):
                    raise ApiError(resp_status)
                sleep(2)
            result = self.matrix_routing_api.get_async_matrix_route_results(result_url)
            return MatrixRoutingResponse.new(result)
        else:
            resp = self.matrix_routing_api.matrix_route(
                origins=origins,
                region_definition=region_definition,
                destinations=destinations,
                profile=profile,
                departure_time=departure_time,
                routing_mode=routing_mode,
                transport_mode=transport_mode,
                avoid_features=avoid_features,
                avoid_areas=avoid_areas,
                truck=truck,
                matrix_attributes=matrix_attributes,
            )
            return MatrixRoutingResponse.new(resp)



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""This module contains classes for accessing `HERE Autosuggest API <https://developer.here.com/documentation/geocoding-search-api/dev_guide/topics/endpoint-autosuggest-brief.html>`_.
"""  # noqa E501

from typing import Dict, List, Optional, Tuple

from here_location_services.config.autosuggest_config import SearchCircle
from here_location_services.platform.auth import Auth

from .apis import Api
from .exceptions import ApiError


class AutosuggestApi(Api):
    """A class for accessing HERE Autosuggest API."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        auth: Optional[Auth] = None,
        proxies: Optional[dict] = None,
        country: str = "row",
    ):
        super().__init__(api_key, auth=auth, proxies=proxies, country=country)
        self._base_url = f"https://autosuggest.search.{self._get_url_string()}"

    def get_autosuggest(
        self,
        query: str,
        at: Optional[List] = None,
        search_in_circle: Optional[SearchCircle] = None,
        search_in_bbox: Optional[Tuple] = None,
        in_country: Optional[List[str]] = None,
        limit: Optional[int] = 20,
        terms_limit: Optional[int] = None,
        lang: Optional[List[str]] = None,
        political_view: Optional[str] = None,
        show: Optional[List[str]] = None,
    ):
        """Suggest address or place candidates based on an incomplete or misspelled query

        :param query: A string for free-text query. Example: res, rest
        :param at: Specify the center of the search context expressed as list of coordinates
            One of `at`, `search_in_circle` or `search_in_bbox` is required.
            Parameters "at", "search_in_circle" and "search_in_bbox" are mutually exclusive. Only
            one of them is allowed.
        :param search_in_circle: Search within a circular geographic area provided as
            latitude, longitude, and radius (in meters)
        :param search_in_bbox: Search within a rectangular bounding box geographic area provided
            as tuple of west longitude, south latitude, east longitude, north latitude
        :param in_country: Search within a specific or multiple countries provided as
            comma-separated ISO 3166-1 alpha-3 country codes. The country codes are to be
            provided in all uppercase. Must be accompanied by exactly one of
            `at`, `search_in_circle` or `search_in_bbox`.
        :param limit: An integer specifiying maximum number of results to be returned.
        :param terms_limit: An integer specifiying maximum number of Query Terms Suggestions
            to be returned.
        :param lang: List of strings to select the language to be used for result rendering
            from a list of BCP 47 compliant language codes.
        :param political_view: Toggle the political view.
        :param show: Select additional fields to be rendered in the response. Please note
            that some of the fields involve additional webservice calls and can increase
            the overall response time.
        :return: :class:`requests.Response` object.
        :raises ApiError: If ``status_code`` of API response is not 200.

        """
        path = "v1/autosuggest"
        url = f"{self._base_url}/{path}"
        params: Dict[str, str] = {
            "q": query,
        }
        if at:
            params["at"] = ",".join([str(i) for i in at])
        if in_country:
            params["in"] = "countryCode:" + ",".join([str(i) for i in in_country])
        if search_in_circle:
            params["in"] = (
                "circle:"
                + str(search_in_circle.lat)
                + ","
                + str(search_in_circle.lng)
                + ";r="
                + str(search_in_circle.radius)
            )
        if limit:
            params["limit"] = str(limit)
        if terms_limit:
            params["termsLimit"] = str(terms_limit)
        if lang:
            params["lang"] = ",".join([str(i) for i in lang])
        if political_view:
            params["politicalView"] = political_view
        if show:
            params["show"] = ",".join([str(i) for i in show])
        if search_in_bbox:
            params["in"] = "bbox:" + ",".join([str(i) for i in search_in_bbox])
        resp = self.get(url, params=params, proxies=self.proxies)
        if resp.status_code == 200:
            return resp
        else:
            raise ApiError(resp)



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""This module defines all the configs which will be required as inputs to matrix Routing API."""
import json
from typing import Dict, List

from .base_config import Bunch


class CircleRegion:
    """A class to define attributes of Circle Region for Matrix Routing API."""

    def __init__(self, center: Dict, radius: int):
        self.type = "circle"
        self.center = center
        self.radius = radius

    def __repr__(self):
        """Return string representation of this instance."""
        return json.dumps(self.__dict__)


class BoundingBoxRegion:
    """A class to define attributes of BoundingBox Region for Matrix Routing API."""

    def __init__(self, north: float, south: float, west: float, east: float):
        self.type = "boundingBox"
        self.north = north
        self.south = south
        self.west = west
        self.east = east

    def __repr__(self):
        """Return string representation of this instance."""
        return json.dumps(self.__dict__)


class PolygonRegion:
    """A class to define attributes of Polygon Region for Matrix Routing API."""

    def __init__(self, outer: List):
        self.type = "polygon"
        self.outer = outer

    def __repr__(self):
        """Return string representation of this instance."""
        return json.dumps(self.__dict__)


class AutoCircleRegion:
    """A class to define attributes of AutoCircle Region for Matrix Routing API."""

    def __init__(self, margin: int = 10000):
        self.type = "autoCircle"
        self.margin = margin

    def __repr__(self):
        """Return string representation of this instance."""
        return json.dumps(self.__dict__)


class WorldRegion:
    """A class to define attributes of World Region for Matrix Routing API."""

    def __init__(self):
        self.type = "world"


class Profile(Bunch):
    """A class to define constant values for ``Profile`` of Matrix Routing API.

    A profile ID enables the calculation of matrices with routes of arbitrary length.
    All profiles explicitly set departureTime to any and require that the obligatory request
    parameter REGION_DEFINITION_TYPE is set to world.

    .. csv-table:: Profiles
        :header: "Profile ID", "Description"
        :widths: 10, 30

        "carFast", "Car with fast routing mode"
        "carShort", "Car with short routing mode"
        "truckFast", "Truck with fast routing mode"
        "pedestrian", "Pedestrian transport mode"
        "bicycle", "Bicycle transport mode"
    """


#: Use this config for profile of matrix Routing API.
#: Example: for ``carFast`` profile  use ``PROFILE.carFast``.
PROFILE = Profile(
    **{
        "carFast": "carFast",
        "carShort": "carShort",
        "truckFast": "truckFast",
        "pedestrian": "pedestrian",
        "bicycle": "bicycle",
    }
)


class MatrixAttributes(Bunch):
    """A class to define constant values for ``matrixAttributes`` of Matrix Routing API.

    MatrixAttributes: ``travelTimes``, ``distances``.
    """


#: Use this config for matrix_attributes of matrix Routing API.
#: Example: for ``travelTimes`` matrix_attributes  use ``MATRIX_ATTRIBUTES.travelTimes``.
MATRIX_ATTRIBUTES = MatrixAttributes(**{"travelTimes": "travelTimes", "distances": "distances"})


class AvoidFeatures(Bunch):
    """A class to define constant values for features to avoid during matrix route calculation."""


#: Use this config for avoid_features of matrix Routing API.
#: Example: for ``tollRoad`` avoid_features  use ``AVOID_FEATURES.tollRoad``.
AVOID_FEATURES = AvoidFeatures(
    **{
        "tollRoad": "tollRoad",
        "controlledAccessHighway": "controlledAccessHighway",
        "ferry": "ferry",
        "tunnel": "tunnel",
        "dirtRoad": "dirtRoad",
    }
)


class AvoidBoundingBox:
    """A class to define attributes of Avoid areaBoundingBox for Matrix Routing API."""

    def __init__(self, north: float, south: float, west: float, east: float):
        self.type = "boundingBox"
        self.north = north
        self.south = south
        self.west = west
        self.east = east

    def __repr__(self):
        """Return string representation of this instance."""
        return json.dumps(self.__dict__)



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""This module defines all the configs which will be required as inputs to Isoline routing API."""

from .base_config import Bunch


class IsolineRoutingTransportMode(Bunch):
    """A class to define constant attributes for mode of transport to be used for the
    calculation of the route.

    * ``car``
    * ``truck``
    * ``pedestrian``
    """


transport_mode = {
    "car": "car",
    "truck": "truck",
    "pedestrian": "pedestrian",
}

#: Use this config for transport_mode of isoline routing API.
#: Example: for ``car`` transport_mode use ``ISOLINE_ROUTING_TRANSPORT_MODE.car``.
ISOLINE_ROUTING_TRANSPORT_MODE = IsolineRoutingTransportMode(**transport_mode)


class RangeType(Bunch):
    """A Class to define constant values for specifying the type of range for Isoline Routing API

    ``distance``:
    Units in meters

    ``time``:
    Units in seconds

    ``consumption``:
    Units in Wh
    """


#: Use this config s optimised_for of isoline routing API.
#: Example: for optimising for ``balanced`` mode use ``OPTIMISED_FOR.balanced``.
RANGE_TYPE = RangeType(**{"distance": "distance", "time": "time", "consumption": "consumption"})


class OptimisedFor(Bunch):
    """A Class to define constant values for optimising calculation for Isoline Routing API

    ``quality``:
    Calculation of isoline focuses on quality, that is, the graph used for isoline calculation
    has higher granularity generating an isoline that is more precise.

    ``performance``:
    Calculation of isoline is performance-centric, quality of isoline is reduced to provide
    better performance.

    ``balanced``:
    Calculation of isoline takes a balanced approach averaging between quality and performance.
    """


#: Use this config s optimised_for of isoline routing API.
#: Example: for optimising for ``balanced`` mode use ``OPTIMISED_FOR.balanced``.
OPTIMISED_FOR = OptimisedFor(
    **{"quality": "quality", "performance": "performance", "balanced": "balanced"}
)


class IsolineRoutingAvoidFeatures(Bunch):
    """A class to define values for features to avoid features during isoline calculation."""


#: Use this config for avoid_features of isoline API.
#: Example: for ``tollRoad`` avoid_features  use ``ISOLINE_ROUTING_AVOID_FEATURES.tollRoad``.
ISOLINE_ROUTING_AVOID_FEATURES = IsolineRoutingAvoidFeatures(
    **{
        "tollRoad": "tollRoad",
        "controlledAccessHighway": "controlledAccessHighway",
        "ferry": "ferry",
        "carShuttleTrain": "carShuttleTrain",
        "tunnel": "tunnel",
        "dirtRoad": "dirtRoad",
        "difficultTurns": "difficultTurns",
    }
)



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""This module contains classes for accessing `HERE Geocoding & Search API <https://developer.here.com/documentation/geocoding-search-api/dev_guide/index.html>`_.
"""  # noqa: E501

from typing import Dict, List, Optional

import requests

from here_location_services.platform.auth import Auth

from .apis import Api
from .exceptions import ApiError


class GeocodingSearchApi(Api):
    """A class for accessing HERE Geocoding & search APIs."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        auth: Optional[Auth] = None,
        proxies: Optional[dict] = None,
        country: str = "row",
    ):
        super().__init__(api_key, auth=auth, proxies=proxies, country=country)
        self._base_url = "https://{0}.search.{1}"

    def get_geocoding(self, query: str, limit: int = 20, lang: str = "en-US") -> requests.Response:
        """
        Get point for given free-form search query.

        See further information here:
        |geocoder|

        .. |geocoder| raw:: html

           <a href="https://developer.here.com/documentation/geocoding-search-api/dev_guide/topics/endpoint-geocode-brief.html" target="_blank">Geocode</a>

        :param query: a string containing the query to make.
        :param limit: An int representing maximum number of results to be returned.
            Default value is 20.
        :param lang: A string to represent language to be used for result rendering from
            a list of BCP47 compliant Language Codes.
        :return: :class:`requests.Response` object.
        :raises ApiError: If ``status_code`` of API response is not 200.
        """  # noqa E501
        path = "/v1/geocode"
        url = self._base_url.format("geocode", self._get_url_string()) + path
        params: dict = dict(q=query, limit=limit, lang=lang)
        resp = self.get(url, params=params, proxies=self.proxies)
        if resp.status_code == 200:
            return resp
        else:
            raise ApiError(resp)

    def get_reverse_geocoding(
        self, lat: float, lng: float, limit: int = 1, lang: str = "en-US"
    ) -> requests.Response:
        """Get address for given latitude and longitude.

        See further information here:
        |reverse_geocode|

        .. |reverse_geocode| raw:: html

           <a href="https://developer.here.com/documentation/geocoding-search-api/dev_guide/topics/endpoint-reverse-geocode-brief.html" target="_blank">Reverse Geocode</a>

        :param lat: A float representing latitude of point.
        :param lng: A float representing longitude of point.
        :param limit: An int representing maximum number of results to be returned.
            Default value is 1.
        :param lang: A string to represent language to be used for result rendering from
            a list of BCP47 compliant Language Codes.
        :return: :class:`requests.Response` object.
        :raises ApiError: If ``status_code`` of API response is not 200.
        """  # noqa E501
        path = "/v1/revgeocode"
        url = self._base_url.format("geocode", self._get_url_string()) + path
        params: dict = dict(at=f"{lat},{lng}", limit=limit, lang=lang)
        resp = self.get(url, params=params, proxies=self.proxies)
        if resp.status_code == 200:
            return resp
        else:
            raise ApiError(resp)

    def get_search_discover(
        self,
        query: str,
        center: Optional[List[float]] = None,
        radius: Optional[int] = None,
        country_codes: Optional[List] = None,
        bounding_box: Optional[List[float]] = None,
        limit: Optional[int] = None,
        lang: Optional[str] = None,
    ) -> requests.Response:
        """Search places using Location Services discover endpoint.

        This method uses location services ``discover`` endpoint to search places based on
        query which is free-form text.

        :param query: A string representing free-test query to search places.
        :param center: A list of latitude and longitude representing the center for
            search query.
        :param radius: A radius in meters along with center for searching places.
        :param country_codes: A list of  ISO 3166-1 alpha-3 country codes.
        :param bounding_box: A bounding box, provided as west longitude, south latitude,
            east longitude, north latitude.
        :param limit: An int representing maximum number of results to be returned.
        :param lang: A string to represent language to be used for result rendering from
            a list of BCP47 compliant Language Codes.
        :return: :class:`requests.Response` object.
        :raises ApiError: If ``status_code`` of API response is not 200.
        """
        path = "/v1/discover"
        url = self._base_url.format("discover", self._get_url_string()) + path
        params: Dict[str, str] = {"q": query}
        if center is not None and radius is not None:
            circle = "circle:" + ",".join([str(x) for x in center])
            params["in"] = circle + f";r={radius}"
        elif center is not None and country_codes:
            params["at"] = ",".join([str(x) for x in center])
            params["in"] = "countryCode:" + ",".join([str(x) for x in country_codes])
        elif bounding_box:
            params["in"] = "bbox:" + ",".join([str(x) for x in bounding_box])
        if lang:
            params["lang"] = lang
        if limit:
            params["limit"] = str(limit)
        resp = self.get(url, params=params, proxies=self.proxies)
        if resp.status_code == 200:
            return resp
        else:
            raise ApiError(resp)

    def get_search_browse(
        self,
        center: List,
        radius: Optional[int] = None,
        country_codes: Optional[List] = None,
        bounding_box: Optional[List[float]] = None,
        categories: Optional[List] = None,
        limit: Optional[int] = None,
        name: Optional[str] = None,
        lang: Optional[str] = None,
    ) -> requests.Response:
        """Get search results for places based on different filters such as categories or name.

        :param center: A list of latitude and longitude representing the center for
            search query.
        :param radius: A radius in meters along with center for searching places.
        :param country_codes: A list of  ISO 3166-1 alpha-3 country codes.
        :param bounding_box: A bounding box, provided as west longitude, south latitude,
            east longitude, north latitude.
        :param categories: A List of strings of category-ids.
        :param limit: An int representing maximum number of results to be returned.
        :param name: A string representing Full-text filter on POI names/titles.
        :param lang: A string to represent language to be used for result rendering from
            a list of BCP47 compliant Language Codes.
        :return: :class:`requests.Response` object.
        :raises ApiError: If ``status_code`` of API response is not 200.
        """
        path = "/v1/browse"
        url = self._base_url.format("browse", self._get_url_string()) + path
        params: Dict[str, str] = {"at": ",".join([str(x) for x in center])}

        if radius is not None:
            params["area"] = "circle:" + ",".join([str(x) for x in center]) + f";r={radius}"
        elif country_codes:
            params["area"] = "countryCode:" + ",".join([str(c) for c in country_codes])
        elif bounding_box:
            params["area"] = "bbox:" + ",".join([str(b) for b in bounding_box])
        if categories:
            params["categories"] = ",".join(categories)
        if name:
            params["name"] = name
        if limit:
            params["limit"] = str(limit)
        if lang:
            params["lang"] = lang
        resp = self.get(url, params=params, proxies=self.proxies)
        if resp.status_code == 200:
            return resp
        else:
            raise ApiError(resp)

    def get_search_lookup(self, location_id: str, lang: Optional[str] = None) -> requests.Response:
        """
        Get search results by providing ``location_id``.

        :param location_id: A string representing id.
        :param lang: A string to represent language to be used for result rendering from
            a list of BCP47 compliant Language Codes.
        :return: :class:`requests.Response` object.
        :raises ApiError: If ``status_code`` of API response is not 200.
        """
        path = "/v1/lookup"
        url = self._base_url.format("lookup", self._get_url_string()) + path
        params: Dict[str, str] = {"id": location_id}
        if lang:
            params["lang"] = lang

        resp = self.get(url, params=params, proxies=self.proxies)
        if resp.status_code == 200:
            return resp
        else:
            raise ApiError(resp)



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains classes for accessing the responses from Location Services RESTful APIs.
"""

import json

import flexpolyline as fp
from geojson import Feature, FeatureCollection, LineString, Point, Polygon
from pandas import DataFrame


class ApiResponse:
    """Base class for all the responses from Location Services RESTful APIs."""

    def __init__(self, **kwargs):
        self._filters = {}
        self.response = None

    def __str__(self):
        return self.as_json_string()

    def as_json_string(self, encoding: str = "utf8"):
        """Return API response as json string."""
        json_string = json.dumps(self.response, sort_keys=True, ensure_ascii=False).encode(
            encoding
        )
        return json_string.decode()

    def to_geojson(self):
        """Return API response as GeoJSON."""
        feature_collection = FeatureCollection([])
        for item in self.response["items"]:
            f = Feature(
                geometry=Point((item["position"]["lng"], item["position"]["lat"])),
                properties=item,
            )
            feature_collection.features.append(f)
        return feature_collection

    @classmethod
    def new(cls, resp):
        """Instantiate a response object from raw response returned by API."""
        json_data = resp.copy()
        obj = cls(**json_data)
        obj.response = resp
        return obj


class GeocoderResponse(ApiResponse):
    """A class representing the Geocoder API response data."""

    def __init__(self, **kwargs):
        super().__init__()
        self._filters = {"items": None}
        for param, default in self._filters.items():
            setattr(self, param, kwargs.get(param, default))


class ReverseGeocoderResponse(ApiResponse):
    """A class representing the Reverse Geocoder API response data."""

    def __init__(self, **kwargs):
        super().__init__()
        self._filters = {"items": None}
        for param, default in self._filters.items():
            setattr(self, param, kwargs.get(param, default))


class IsolineResponse(ApiResponse):
    """A class representing the Reverse Isoline routing API response data."""

    def __init__(self, **kwargs):
        super().__init__()
        self._filters = {
            "departure": None,
            "arrival": None,
            "isolines": None,
            "notices": None,
        }
        for param, default in self._filters.items():
            setattr(self, param, kwargs.get(param, default))

    def to_geojson(self):
        """Return API response as GeoJSON."""
        feature_collection = FeatureCollection([])
        for isoline in self.response["isolines"]:
            for polygon in isoline["polygons"]:
                polyline = polygon["outer"]
                lstring = fp.decode(polyline)
                lstring = [(coord[1], coord[0]) for coord in lstring]
                f = Feature(geometry=Polygon([lstring]), properties={"range": isoline["range"]})
                feature_collection.features.append(f)
        return feature_collection


class DiscoverResponse(ApiResponse):
    """A class representing the search discover API response data."""

    def __init__(self, **kwargs):
        super().__init__()
        self._filters = {"items": None}
        for param, default in self._filters.items():
            setattr(self, param, kwargs.get(param, default))


class BrowseResponse(ApiResponse):
    """A class representing the search browse API response data."""

    def __init__(self, **kwargs):
        super().__init__()
        self._filters = {"items": None}
        for param, default in self._filters.items():
            setattr(self, param, kwargs.get(param, default))


class LookupResponse(ApiResponse):
    """A class representing the search lookup API response data."""

    def __init__(self, **kwargs):
        super().__init__()
        self._filters = {"items": None}
        for param, default in self._filters.items():
            setattr(self, param, kwargs)


class RoutingResponse(ApiResponse):
    """A class representing the search routing API response data."""

    def __init__(self, **kwargs):
        super().__init__()
        self._filters = {"routes": None}
        for param, default in self._filters.items():
            setattr(self, param, kwargs.get(param, default))

    def to_geojson(self):
        """Return API response as GeoJSON."""
        feature_collection = FeatureCollection([])
        for route in self.response["routes"]:
            for section in route["sections"]:
                polyline = section["polyline"]
                lstring = fp.decode(polyline)
                lstring = [(coord[1], coord[0], coord[2]) for coord in lstring]
                f = Feature(geometry=LineString(lstring), properties=section)
                feature_collection.features.append(f)
        return feature_collection


class MatrixRoutingResponse(ApiResponse):
    """A class representing Matrix routing response data."""

    def __init__(self, **kwargs):
        super().__init__()
        self._filters = {"matrix": None}
        for param, default in self._filters.items():
            setattr(self, param, kwargs.get(param, default))

    def to_geojson(self):
        """Return API response as GeoJSON."""
        raise NotImplementedError("This method is not valid for MatrixRoutingResponse.")

    def to_distnaces_matrix(self):
        """Return distnaces matrix in a dataframe."""
        if self.matrix and self.matrix.get("distances"):
            distances = self.matrix.get("distances")
            dest_count = self.matrix.get("numDestinations")
            nested_distances = [
                distances[i : i + dest_count] for i in range(0, len(distances), dest_count)
            ]
            return DataFrame(nested_distances, columns=range(dest_count))

    def to_travel_times_matrix(self):
        """Return travel times matrix in a dataframe."""
        if self.matrix and self.matrix.get("travelTimes"):
            distances = self.matrix.get("travelTimes")
            dest_count = self.matrix.get("numDestinations")
            nested_distances = [
                distances[i : i + dest_count] for i in range(0, len(distances), dest_count)
            ]
            return DataFrame(nested_distances, columns=range(dest_count))


class AutosuggestResponse(ApiResponse):
    """A class representing the Autosuggest API response data."""

    def __init__(self, **kwargs):
        super().__init__()
        self._filters = {
            "items": None,
            "queryTerms": None,
        }
        for param, default in self._filters.items():
            setattr(self, param, kwargs.get(param, default))


class DestinationWeatherResponse(ApiResponse):
    """A class representing the Destination Weather API response data."""

    def __init__(self, **kwargs):
        super().__init__()
        self._filters = {"places": None, "features": None}
        for param, default in self._filters.items():
            setattr(self, param, kwargs.get(param, default))



# Copyright (C) 2019-2021 HERE Europe B.V.
# SPDX-License-Identifier: Apache-2.0

"""This module defines all the configs which will be required as inputs to routing APIs."""

import json
from typing import Optional

from .base_config import Bunch, PlaceOptions, WayPointOptions


class RoutingReturn(Bunch):
    """A class to define constant attributes which are included in routing API's response
    as part of the data representation of route or section.

    * ``polyline`` - Polyline for the route in Flexible Polyline Encoding. Either a 2D polyline
      (without elevation specified) or a 3D polyline with the 3rd dimension type Elevation
      (with elevation specified).

    * ``actions`` - Actions (such as maneuvers or tasks) that must be taken to complete the
      section.

    * instructions - Include instructions in returned actions. Instructions are localized to the
      requested language.

    * ``summary`` - Include a summary for the section.

    * ``travelSummary`` - Include a summary for the travel portion of the section.

    * ``turnByTurnActions`` - Include all information necessary to support turn by turn guidance
      to complete the section.

    * ``mlDuration`` - Use a region-specific machine learning model to calculate route duration.
      Disclaimer: This parameter is currently in beta release, and is therefore subject
      to breaking changes.

    * ``elevation`` - Include elevation information in coordinate and geometry types.
      See e.g. polyline or location.

    * ``routeHandle`` - Encode calculated route and return a handle which can be used with
      routes/{routeHandle} to decode the route at a later point in time.

    * ``incidents`` - Include a list of all incidents applicable to each section.

    Following restrictions apply when specifying return parameter:

    * If ``actions`` is requested, then ``polyline`` must also be requested as well.

    * If ``instructions`` is requested, then ``actions`` must also be requested as well.

    * If ``turnByTurnActions`` is requested, then ``polyline`` must also be requested as well.

    * If at least one attribute is requested within the ``spans`` parameter, then ``polyline``
      must be requested as well.
    """


return_attributes = {
    "polyline": "polyline",
    "actions": "actions",
    "instructions": "instructions",
    "summary": "summary",
    "travelSummary": "travelSummary",
    "mlDuration": "mlDuration",
    "turnByTurnActions": "turnByTurnActions",
    "elevation": "elevation",
    "routeHandle": "routeHandle",
    "passthrough": "passthrough",
    "incidents": "incidents",
}
#: Use this config for return attributes from routing API for param ``return_results``.
#: Example: To return polyline in results use ``ROUTING_RETURN.polyline``.
ROUTING_RETURN = RoutingReturn(**return_attributes)


class RoutingSpans(Bunch):
    """A class to define constant attributes which are included in the response spans.

    ``walkAttributes`` ``streetAttributes`` ``carAttributes`` ``truckAttributes``
    ``scooterAttributes`` ``names`` ``length`` ``duration`` ``baseDuration`` ``countryCode``
    ``functionalClass`` ``routeNumbers`` ``speedLimit`` ``maxSpeed`` ``dynamicSpeedInfo``
    ``segmentId`` ``segmentRef`` ``consumption``.
    """


routing_spans = {
    "walkAttributes": "walkAttributes",
    "streetAttributes": "streetAttributes",
    "carAttributes": "streetAttributes",
    "truckAttributes": "truckAttributes",
    "scooterAttributes": "scooterAttributes",
    "names": "names",
    "length": "length",
    "duration": "duration",
    "baseDuration": "baseDuration",
    "countryCode": "countryCode",
    "functionalClass": "functionalClass",
    "routeNumbers": "routeNumbers",
    "speedLimit": "speedLimit",
    "maxSpeed": "maxSpeed",
    "dynamicSpeedInfo": "dynamicSpeedInfo",
    "segmentId": "segmentId",
    "segmentRef": "segmentRef",
    "consumption": "consumption",
}

#: Use this config for spans of routing API.
#: Example: for ``walkAttributes`` routing_mode use ``ROUTING_SPANS.walkAttributes``.
ROUTING_SPANS = RoutingSpans(**routing_spans)


class RoutingTransportMode(Bunch):
    """A class to define constant attributes for mode of transport to be used for the
    calculation of the route.

    * ``car``
    * ``truck``
    * ``pedestrian``
    * ``bicycle``
    * ``scooter``
    """


transport_mode = {
    "car": "car",
    "truck": "truck",
    "pedestrian": "pedestrian",
    "bicycle": "bicycle",
    "scooter": "scooter",
}

#: Use this config for transport_mode of routing API.
#: Example: for ``car`` transport_mode use ``ROUTING_TRANSPORT_MODE.car``.
ROUTING_TRANSPORT_MODE = RoutingTransportMode(**transport_mode)


class RouteCourse(Bunch):
    """A class to define constant attributes for Course option."""


route_course = {"east": 90, "south": 180, "west": 270, "north": 360}

ROUTE_COURSE = RouteCourse(**route_course)


class RouteMatchSideOfStreet(Bunch):
    """A class to define constant attribuites for ``matchSideOfStreet``."""


match_sideof_street = {"always": "always", "onlyIfDivided": "onlyIfDivided"}

#: Use this config for transport_mode of routing API.
#: Example: for ``car`` transport_mode use ``ROUTING_TRANSPORT_MODE.car``.
ROUTE_MATCH_SIDEOF_STREET = RouteMatchSideOfStreet(**match_sideof_street)


class AvoidFeatures(Bunch):
    """A class to define constant values for features to avoid during route calculation."""


#: Use this config for match_sideof_street for PlaceOptions.
#: Example: for match_sideof_street ``always`` use ``ROUTE_MATCH_SIDEOF_STREET.always``.
AVOID_FEATURES = AvoidFeatures(
    **{
        "seasonalClosure": "seasonalClosure",
        "tollRoad": "tollRoad",
        "controlledAccessHighway": "controlledAccessHighway",
        "ferry": "ferry",
        "carShuttleTrain": "carShuttleTrain",
        "tunnel": "tunnel",
        "dirtRoad": "dirtRoad",
        "difficultTurns": "difficultTurns",
    }
)


class Scooter:
    """A class to define attributes specific for the scooter route.

    Scooter specific parameters:

    allowHighway: Specifies whether the scooter is allowed on the highway or not.
    This parameter is optional. If not provided, then by default scooter is not allowed to use
    highway. There is a similar parameter ``avoid[features]=controlledAccessHighway`` to disallow
    highway usage. avoid[features] takes precedence so if this parameter is also used then
    scooters are not allowed to use highways even if allowHighway is used with value as true.
    Possible values:
    true: scooter is allowed to use the highway.
    false: scooter is not allowed to use the highway.
    """

    def __init__(self, allow_highway: bool):
        self.allowHighway = allow_highway

    def __repr__(self):
        """Return string representation of this instance."""
        return json.dumps(self.__dict__)


class Via:
    """A class to define ``via`` waypoint.

    A via waypoint is a location between origin and destination. The route will do a stop at the
    via waypoint.
    """

    def __init__(
        self,
        lat: float,
        lng: float,
        place_options: Optional[PlaceOptions] = None,
        waypoint_options: Optional[WayPointOptions] = None,
    ):
        self.lat = lat
        self.lng = lng
        self.place_options = place_options
        self.waypoint_options = waypoint_options
